__all__ = ('SlasherSnapshotType', )


from scarletio import RichAttributeErrorBaseType, copy_docs

from ..plugin_loader import BaseSnapshotType, PLUGIN_LOADER
from ..plugin_loader.snapshot.helpers import (
    _get_list_difference, _get_set_difference, _merge_list_groups, _merge_set_groups
)

from .slasher import Slasher
from .utils import RUNTIME_SYNC_HOOKS, SYNC_ID_NON_GLOBAL


class ApplicationCommandDifference(RichAttributeErrorBaseType):
    """
    Snapshot for application commands.
    
    Attributes
    ----------
    added_application_commands : `None`, `list` of ``CommandBaseApplicationCommand``
        The added application commands.
    removed_application_commands : `None`, `list` of ``CommandBaseApplicationCommand``
        The removed application commands.
    """
    __slots__ = ('added_application_commands', 'removed_application_commands',)
    
    def __new__(cls):
        """
        Creates a new application command snapshot.
        """
        self = object.__new__(cls)
        self.added_application_commands = None
        self.removed_application_commands = None
        return self
    
    def __repr__(self):
        """Returns the application command difference."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        added_application_commands = self.added_application_commands
        if (added_application_commands is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' added_application_commands=')
            repr_parts.append(repr(added_application_commands))
        
        removed_application_commands = self.removed_application_commands
        if (removed_application_commands is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' removed_application_commands=')
            repr_parts.append(repr(removed_application_commands))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __bool__(self):
        """Returns whether the difference contains any changes."""
        if (self.added_application_commands is not None):
            return True
        
        if (self.removed_application_commands is not None):
            return True
        
        return False
    
    
    def add(self, application_command):
        """
        Adds an application command to self.
        
        Parameters
        ----------
        application_command : ``CommandBaseApplicationCommand``
            The application command to add.
        """
        added_application_commands = self.added_application_commands
        if (added_application_commands is None):
            added_application_commands = []
            self.added_application_commands = added_application_commands
        
        removed_application_commands = self.removed_application_commands
        if (removed_application_commands is not None):
            try:
                removed_application_commands.remove(application_command)
            except ValueError:
                pass
            else:
                if not removed_application_commands:
                    self.removed_application_commands = None
        
        added_application_commands.append(application_command)
    
    
    def remove(self, application_command):
        """
        Removes the application command from self.
        
        Parameters
        ----------
        application_command : ``CommandBaseApplicationCommand``
            The application command to remove.
        """
        added_application_commands = self.added_application_commands
        if (added_application_commands is not None):
            try:
                added_application_commands.remove(application_command)
            except ValueError:
                pass
            else:
                if not added_application_commands:
                    self.added_application_commands = None
        
        removed_application_commands = self.removed_application_commands
        if (removed_application_commands is None):
            removed_application_commands = []
            self.removed_application_commands = removed_application_commands
        
        removed_application_commands.append(application_command)
    
    
    def iter_added_application_commands(self):
        """
        Iterates over the added application commands of the application command difference.
        
        This method is an iterable generator.
        
        Yields
        ------
        added_application_command : ``CommandBaseApplicationCommand``
        """
        added_application_commands = self.added_application_commands
        if (added_application_commands is not None):
            yield from added_application_commands
    
    
    def iter_removed_application_commands(self):
        """
        Iterates over the removed application commands of the application command difference.
        
        This method is an iterable generator.
        
        Yields
        ------
        removed_application_command : ``CommandBaseApplicationCommand``
        """
        removed_application_commands = self.removed_application_commands
        if (removed_application_commands is not None):
            yield from removed_application_commands
    
    
    def copy(self):
        """
        Copies the application command difference.
        
        Returns
        -------
        new : ``ApplicationCommandDifference``
        """
        added_application_commands = self.added_application_commands
        if (added_application_commands is not None):
            added_application_commands = added_application_commands.copy()
        
        removed_application_commands = self.removed_application_commands
        if (removed_application_commands is not None):
            removed_application_commands = removed_application_commands.copy()
        
        new = object.__new__(type(self))
        new.added_application_commands = added_application_commands
        new.removed_application_commands = removed_application_commands
        return new
    
    
    def revert_copy(self):
        """
        Revert copies the application command difference.
        
        Returns
        -------
        new : ``ApplicationCommandDifference``
        """
        added_application_commands = self.added_application_commands
        if (added_application_commands is not None):
            added_application_commands = added_application_commands.copy()
        
        removed_application_commands = self.removed_application_commands
        if (removed_application_commands is not None):
            removed_application_commands = removed_application_commands.copy()
        
        new = object.__new__(type(self))
        new.added_application_commands = removed_application_commands
        new.removed_application_commands = added_application_commands
        return new
    
    
    def _subtract(self, other):
        """
        Subtracts other from self.
        
        Parameters
        ----------
        other : ``ApplicationCommandDifference``
            The other difference to subtract from self.
        
        Returns
        -------
        new : ``ApplicationCommandDifference``
        """
        added_application_commands, removed_application_commands = _get_list_difference(
            _merge_list_groups(self.added_application_commands, other.removed_application_commands),
            _merge_list_groups(other.added_application_commands, self.removed_application_commands),
        )
        
        new = object.__new__(type(self))
        new.added_application_commands = added_application_commands
        new.removed_application_commands = removed_application_commands
        return new
    
    
    def __sub__(self, other):
        """Subtracts other from self."""
        if other is None:
            return self.copy()
        
        if type(self) is type(other):
            return self._subtract(other)
        
        return NotImplemented
    
    
    def __rsub__(self, other):
        """subtracts other from self."""
        if other is None:
            return self.revert_copy()
        
        if type(self) is type(other):
            return other._subtract(self)
        
        return NotImplemented


class PermissionOverwriteDifference(RichAttributeErrorBaseType):
    """
    Snapshot for permission overwrites.
    
    Attributes
    ----------
    added_permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The added permission overwrites.
    removed_permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The removed permission overwrites.
    """
    __slots__ = ('added_permission_overwrites', 'removed_permission_overwrites',)
    
    def __new__(cls):
        """
        Creates a new permission overwrite snapshot.
        """
        self = object.__new__(cls)
        self.added_permission_overwrites = None
        self.removed_permission_overwrites = None
        return self
    
    def __repr__(self):
        """Returns the permission overwrite difference."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        added_permission_overwrites = self.added_permission_overwrites
        if (added_permission_overwrites is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' added_permission_overwrites=')
            repr_parts.append(repr(added_permission_overwrites))
        
        removed_permission_overwrites = self.removed_permission_overwrites
        if (removed_permission_overwrites is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' removed_permission_overwrites=')
            repr_parts.append(repr(removed_permission_overwrites))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __bool__(self):
        """Returns whether the difference contains any changes."""
        if (self.added_permission_overwrites is not None):
            return True
        
        if (self.removed_permission_overwrites is not None):
            return True
        
        return False
    
    
    def add(self, permission_overwrite):
        """
        Adds an permission overwrite to self.
        
        Parameters
        ----------
        permission_overwrite : ``ApplicationCommandPermissionOverwrite``
            The permission overwrite to add.
        """
        added_permission_overwrites = self.added_permission_overwrites
        if (added_permission_overwrites is None):
            added_permission_overwrites = []
            self.added_permission_overwrites = added_permission_overwrites
        
        removed_permission_overwrites = self.removed_permission_overwrites
        if (removed_permission_overwrites is not None):
            try:
                removed_permission_overwrites.remove(permission_overwrite)
            except ValueError:
                pass
            else:
                if not removed_permission_overwrites:
                    self.removed_permission_overwrites = None
        
        added_permission_overwrites.append(permission_overwrite)
    
    
    def remove(self, permission_overwrite):
        """
        Removes the permission overwrite from self.
        
        Parameters
        ----------
        permission_overwrite : ``ApplicationCommandPermissionOverwrite``
            The permission overwrite to remove.
        """
        added_permission_overwrites = self.added_permission_overwrites
        if (added_permission_overwrites is not None):
            try:
                added_permission_overwrites.remove(permission_overwrite)
            except ValueError:
                pass
            else:
                if not added_permission_overwrites:
                    self.added_permission_overwrites = None
        
        removed_permission_overwrites = self.removed_permission_overwrites
        if (removed_permission_overwrites is None):
            removed_permission_overwrites = []
            self.removed_permission_overwrites = removed_permission_overwrites
        
        removed_permission_overwrites.append(permission_overwrite)
    
    
    def iter_added_permission_overwrites(self):
        """
        Iterates over the added permission overwrites of the permission overwrite difference.
        
        This method is an iterable generator.
        
        Yields
        ------
        added_permission_overwrite : ``ApplicationCommandPermissionOverwrite``
        """
        added_permission_overwrites = self.added_permission_overwrites
        if (added_permission_overwrites is not None):
            yield from added_permission_overwrites
    
    
    def iter_removed_permission_overwrites(self):
        """
        Iterates over the removed permission overwrites of the permission overwrite difference.
        
        This method is an iterable generator.
        
        Yields
        ------
        removed_permission_overwrite : ``ApplicationCommandPermissionOverwrite``
        """
        removed_permission_overwrites = self.removed_permission_overwrites
        if (removed_permission_overwrites is not None):
            yield from removed_permission_overwrites
    
    
    def copy(self):
        """
        Copies the permission overwrite difference.
        
        Returns
        -------
        new : ``PermissionOverwriteDifference``
        """
        added_permission_overwrites = self.added_permission_overwrites
        if (added_permission_overwrites is not None):
            added_permission_overwrites = added_permission_overwrites.copy()
        
        removed_permission_overwrites = self.removed_permission_overwrites
        if (removed_permission_overwrites is not None):
            removed_permission_overwrites = removed_permission_overwrites.copy()
        
        new = object.__new__(type(self))
        new.added_permission_overwrites = added_permission_overwrites
        new.removed_permission_overwrites = removed_permission_overwrites
        return new
    
    
    def revert_copy(self):
        """
        Revert copies the permission overwrite difference.
        
        Returns
        -------
        new : ``PermissionOverwriteDifference``
        """
        added_permission_overwrites = self.added_permission_overwrites
        if (added_permission_overwrites is not None):
            added_permission_overwrites = added_permission_overwrites.copy()
        
        removed_permission_overwrites = self.removed_permission_overwrites
        if (removed_permission_overwrites is not None):
            removed_permission_overwrites = removed_permission_overwrites.copy()
        
        new = object.__new__(type(self))
        new.added_permission_overwrites = removed_permission_overwrites
        new.removed_permission_overwrites = added_permission_overwrites
        return new
    
    
    def _subtract(self, other):
        """
        Subtracts other from self.
        
        Parameters
        ----------
        other : ``PermissionOverwriteDifference``
            The other difference to subtract from self.
        
        Returns
        -------
        new : ``PermissionOverwriteDifference``
        """
        added_permission_overwrites, removed_permission_overwrites = _get_list_difference(
            _merge_list_groups(self.added_permission_overwrites, other.removed_permission_overwrites),
            _merge_list_groups(other.added_permission_overwrites, self.removed_permission_overwrites),
        )
        
        new = object.__new__(type(self))
        new.added_permission_overwrites = added_permission_overwrites
        new.removed_permission_overwrites = removed_permission_overwrites
        return new
    
    
    def __sub__(self, other):
        """Subtracts other from self."""
        if other is None:
            return self.copy()
        
        if type(self) is type(other):
            return self._subtract(other)
        
        return NotImplemented
    
    
    def __rsub__(self, other):
        """subtracts other from self."""
        if other is None:
            return self.revert_copy()
        
        if type(self) is type(other):
            return other._subtract(self)
        
        return NotImplemented


class SlasherSnapshotType(BaseSnapshotType):
    """
    Slasher extension snapshot.
    
    Attributes
    ----------
    added_component_commands : `None`, `set` of ``ComponentCommand``
        The added component commands
    added_form_submit_commands : `None, `set` of ``FormSubmitCommand``
        The added form submit commands.
    application_command_differences_by_guild_id : `dict` of (`int`, ``ApplicationCommandDifference``)
        The added & removed application commands by guild id.
    permission_overwrite_difference_by_guild_id : `dict` of (`int`, ``PermissionOverwriteDifference``)
        The added & removed application command permission overwrites by guild id.
    removed_component_commands : `None`, `set` of ``ComponentCommand``
        The removed component commands
    removed_form_submit_commands : `None, `set` of ``FormSubmitCommand``
        The removed form submit commands.
    """
    __slots__ = (
        'added_component_commands', 'added_form_submit_commands', 'application_command_differences_by_guild_id',
        'permission_overwrite_difference_by_guild_id', 'removed_component_commands', 'removed_form_submit_commands',
    )
    
    
    @copy_docs(BaseSnapshotType.__new__)
    def __new__(cls, client):
        application_command_differences_by_guild_id = {}
        permission_overwrite_difference_by_guild_id = {}
        
        slasher = getattr(client, 'slasher', None)
        if (slasher is None) or (not isinstance(slasher, Slasher)):
            added_component_commands = None
            added_form_submit_commands = None
        
        else:
            command_states = slasher._command_states
            
            for guild_id, command_state in command_states.items():
                application_command_difference = ApplicationCommandDifference()
                
                if guild_id == SYNC_ID_NON_GLOBAL:
                    active_commands = command_state._active
                    if (active_commands is not None):
                        for command in active_commands:
                            application_command_difference.add(command)
                    
                else:
                    changes = command_state._changes
                    if (changes is not None):
                        for added, command in changes:
                            if added:
                                application_command_difference.add(command)
                            else:
                                application_command_difference.remove(command)
                
                if application_command_difference:
                    application_command_differences_by_guild_id[guild_id] = application_command_difference
            
            added_component_commands = slasher._component_commands
            if added_component_commands:
                added_component_commands = added_component_commands.copy()
            else:
                added_component_commands = None
            
            added_form_submit_commands = slasher._form_submit_commands
            if added_form_submit_commands:
                added_form_submit_commands = added_form_submit_commands.copy()
            else:
                added_form_submit_commands = None
            
            # permission_overwrite_difference_by_guild_id
            guild_level_permission_overwrites = slasher._guild_level_permission_overwrites
            if guild_level_permission_overwrites is not None:
                for guild_id, permission_overwrites in guild_level_permission_overwrites.items():
                    permission_overwrite_difference = PermissionOverwriteDifference()
                    for permission_overwrite in permission_overwrites:
                        permission_overwrite_difference.add(permission_overwrite)
                    
                    if permission_overwrite_difference:
                        permission_overwrite_difference_by_guild_id[guild_id] = permission_overwrite_difference
        
        
        self = BaseSnapshotType.__new__(cls, client)
        
        self.added_component_commands = added_component_commands
        self.added_form_submit_commands = added_form_submit_commands
        self.application_command_differences_by_guild_id = application_command_differences_by_guild_id
        self.permission_overwrite_difference_by_guild_id = permission_overwrite_difference_by_guild_id
        self.removed_component_commands = None
        self.removed_form_submit_commands = None
        
        return self
    
    
    @copy_docs(BaseSnapshotType.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        client = self.client
        if (client is not None):
            repr_parts.append(' of ')
            repr_parts.append(repr(client))
        
        application_command_differences_by_guild_id = self.application_command_differences_by_guild_id
        if application_command_differences_by_guild_id:
            repr_parts.append(', application_command_differences_by_guild_id = ')
            repr_parts.append(repr(application_command_differences_by_guild_id))
        
        permission_overwrite_difference_by_guild_id = self.permission_overwrite_difference_by_guild_id
        if permission_overwrite_difference_by_guild_id:
            repr_parts.append(', permission_overwrite_difference_by_guild_id = ')
            repr_parts.append(repr(permission_overwrite_difference_by_guild_id))
        
        added_component_commands = self.added_component_commands
        if (added_component_commands is not None):
            repr_parts.append(', added_component_commands = ')
            repr_parts.append(repr(added_component_commands))
            
        removed_component_commands = self.removed_component_commands
        if (removed_component_commands is not None):
            repr_parts.append(', removed_component_commands = ')
            repr_parts.append(repr(removed_component_commands))
        
        added_form_submit_commands = self.added_form_submit_commands
        if (added_form_submit_commands is not None):
            repr_parts.append(', added_form_submit_commands = ')
            repr_parts.append(repr(added_form_submit_commands))
        
        removed_form_submit_commands = self.removed_form_submit_commands
        if (removed_form_submit_commands is not None):
            repr_parts.append(', removed_form_submit_commands = ')
            repr_parts.append(repr(removed_form_submit_commands))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(BaseSnapshotType._extract)
    def _extract(self, other):
        self_application_command_differences_by_guild_id = self.application_command_differences_by_guild_id
        other_application_command_differences_by_guild_id = other.application_command_differences_by_guild_id
        
        new_application_command_differences_by_guild_id = {}
        
        for guild_id in {
            *self_application_command_differences_by_guild_id.keys(),
            *other_application_command_differences_by_guild_id.keys(),
        }:
            new_application_command_difference = (
                self_application_command_differences_by_guild_id.get(guild_id, None) -
                other_application_command_differences_by_guild_id.get(guild_id, None)
            )
            
            if new_application_command_difference:
                new_application_command_differences_by_guild_id[guild_id] = new_application_command_difference
        
        
        self_permission_overwrite_difference_by_guild_id = self.permission_overwrite_difference_by_guild_id
        other_permission_overwrite_difference_by_guild_id = other.permission_overwrite_difference_by_guild_id
        
        new_permission_overwrite_difference_by_guild_id = {}
        
        for guild_id in {
            *self_permission_overwrite_difference_by_guild_id.keys(),
            *other_permission_overwrite_difference_by_guild_id.keys(),
        }:
            new_application_command_difference = (
                self_permission_overwrite_difference_by_guild_id.get(guild_id, None) -
                other_permission_overwrite_difference_by_guild_id.get(guild_id, None)
            )
            
            if new_application_command_difference:
                new_permission_overwrite_difference_by_guild_id[guild_id] = new_application_command_difference
        
        
        added_component_commands, removed_component_commands = _get_set_difference(
            _merge_set_groups(self.added_component_commands, other.removed_component_commands),
            _merge_set_groups(other.added_component_commands, self.removed_component_commands),
        )
        
        added_form_submit_commands, removed_form_submit_commands = _get_set_difference(
            _merge_set_groups(self.added_form_submit_commands, other.removed_form_submit_commands),
            _merge_set_groups(other.added_form_submit_commands, self.removed_form_submit_commands),
        )
        
        new = BaseSnapshotType._extract(self, other)
        
        new.added_component_commands = added_component_commands
        new.added_form_submit_commands = added_form_submit_commands
        new.application_command_differences_by_guild_id = new_application_command_differences_by_guild_id
        new.permission_overwrite_difference_by_guild_id = new_permission_overwrite_difference_by_guild_id
        new.removed_component_commands = removed_component_commands
        new.removed_form_submit_commands = removed_form_submit_commands
        
        client = self.client
        if client.running and client.application.id:
            slasher = getattr(client, 'slasher', None)
            if (slasher is not None):
                PLUGIN_LOADER.add_done_callback_unique(slasher.sync)
        
        return new
    
    
    @copy_docs(BaseSnapshotType.revert)
    def revert(self):
        BaseSnapshotType.revert(self)
        
        client = self.client
        if client is None:
            return

        slasher = getattr(client, 'slasher', None)
        if (slasher is None) or (not isinstance(slasher, Slasher)):
            return
        
        for application_command_difference in self.application_command_differences_by_guild_id.values():
            for application_command in application_command_difference.iter_added_application_commands():
                slasher._remove_application_command(application_command)
            
            for application_command in application_command_difference.iter_removed_application_commands():
                slasher._add_application_command(application_command)
        
        for guild_id, permission_overwrite_difference in self.permission_overwrite_difference_by_guild_id.items():
            for permission_overwrite in permission_overwrite_difference.iter_added_permission_overwrites():
                slasher._remove_permission_overwrite_for_guild(guild_id, permission_overwrite)
            
            for permission_overwrite in permission_overwrite_difference.iter_removed_permission_overwrites():
                slasher._add_permission_overwrites_for_guild(guild_id, permission_overwrite)
        
        
        added_component_commands = self.added_component_commands
        if (added_component_commands is not None):
            for component_command in added_component_commands:
                slasher._remove_component_command(component_command)
        
        
        removed_component_commands = self.removed_component_commands
        if (removed_component_commands is not None):
            for component_command in removed_component_commands:
                slasher._add_component_command(component_command)
        
        added_form_submit_commands = self.added_form_submit_commands
        if (added_form_submit_commands is not None):
            for form_submit_command in added_form_submit_commands:
                slasher._remove_form_submit_command(form_submit_command)
        
        removed_form_submit_commands = self.removed_form_submit_commands
        if (removed_form_submit_commands is not None):
            for form_submit_command in removed_form_submit_commands:
                slasher._add_form_submit_command(form_submit_command)
    
    
    @copy_docs(BaseSnapshotType.__bool__)
    def __bool__(self):
        if (self.added_component_commands is not None):
            return True
        
        if (self.added_form_submit_commands is not None):
            return True
        
        if self.application_command_differences_by_guild_id:
            return True
        
        if self.permission_overwrite_difference_by_guild_id:
            return True
        
        if (self.removed_component_commands is not None):
            return True
        
        if (self.removed_form_submit_commands is not None):
            return True
        
        return False


def runtime_sync_hook_is_executing_extension(client):
    """
    Runtime sync hook to check whether a slash command should be registered and synced instantly when added or removed.
    
    Parameters
    ----------
    client : ``Client``
        The respective client of the ``Slasher``.
    """
    return not PLUGIN_LOADER.is_processing_plugin()

RUNTIME_SYNC_HOOKS.append(runtime_sync_hook_is_executing_extension)



__all__ = ()

from ...field_parsers import entity_id_array_parser_factory, nullable_string_parser_factory
from ...field_putters import optional_entity_id_array_optional_putter_factory, nullable_string_putter_factory
from ...field_validators import entity_id_array_validator_factory, nullable_string_validator_factory
from ...user import ClientUserBase

# location

parse_location = nullable_string_parser_factory('location')
put_location_into = nullable_string_putter_factory('location')
validate_location = nullable_string_validator_factory('location', 0, 1024)

# speaker_ids

parse_speaker_ids = entity_id_array_parser_factory('speaker_ids')
put_speaker_ids_into = optional_entity_id_array_optional_putter_factory('speaker_ids')
validate_speaker_ids = entity_id_array_validator_factory('speaker_ids', ClientUserBase)



__all__ = ('AuditLogRole',)

from ..bases import DiscordEntity
from ..role import Role


class AuditLogRole(DiscordEntity):
    """
    Represents a role object received with audit logs.
    
    Parameters
    ----------
    id : `int`
        The role's identifier.
    name : `str`
        The role's name.
    """
    __slots__ = ('id', 'name')
    
    def __new__(cls, data):
        """
        Creates a new audit log role.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            partial role data.
        """
        role_id = int(data['id'])
        role_name = data.get('name', '')
        
        self = object.__new__(cls)
        self.id = role_id
        self.name = role_name
        return self
    
    
    def __repr__(self):
        """Returns the role's representation."""
        return f'<{self.__class__.__name__} id = {self.id!r}, name = {self.name!r}>'
    
    
    @property
    def entity(self):
        """
        Resolves the audit log role's entity.
        
        If the entity is not cached creates a new.
        
        Returns
        -------
        role : ``Role``
        """
        return Role.precreate(self.id, name = self.name)



__all__ = ('GuildWidgetUser',)

import warnings

from ...bases import DiscordEntity
from ...user import Status

from .fields import (
    parse_activity_name, parse_avatar_url, parse_discriminator, parse_id, parse_name, parse_status,
    put_activity_name_into, put_avatar_url_into, put_discriminator_into, put_id_into, put_name_into, put_status_into,
    validate_activity_name, validate_avatar_url, validate_discriminator, validate_id, validate_name, validate_status
)


class GuildWidgetUser(DiscordEntity):
    """
    Represents an user object sent with a ``GuildWidget``'s data.
    
    Attributes
    ----------
    activity_name : `None`, `str`
        The guild widget user's activity's name if applicable.
    avatar_url : `str`
        The guild widget user's avatar url.
    discriminator : `int`
        The guild widget user's discriminator.
    id : `int`
        The unique identifier number of the guild widget user. Can be between `0` and `99`.
    name : `str`
        The guild widget user's name.
    status : ``Status``
        The guild widget user's status.
    """
    __slots__ = ('activity_name', 'avatar_url', 'discriminator', 'name', 'status')
    
    
    def __new__(
        cls,
        *,
        activity_name = ...,
        avatar_url = ...,
        discriminator = ...,
        name = ...,
        status = ...,
        user_id = ...,
    ):
        """
        Creates a new guild widget user with the given parameters.
        
        Parameters
        ----------
        activity_name : `None`, `str`, Optional (Keyword only)
            The guild widget user's activity's name if applicable.
        avatar_url : `str`, Optional (Keyword only)
            The guild widget user's avatar url.
        discriminator : `int`, Optional (Keyword only)
            The guild widget user's discriminator.
        name : `str`, Optional (Keyword only)
            The guild widget user's name.
        status : ``Status``, `str`, Optional (Keyword only)
            The guild widget user's status.
        user_id : `int`, Optional (Keyword only)
            The unique identifier number of the guild widget user. Can be between `0` and `99`.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # activity_name
        if activity_name is ...:
            activity_name = None
        else:
            activity_name = validate_activity_name(activity_name)
        
        # avatar_url
        if avatar_url is ...:
            avatar_url = ''
        else:
            avatar_url = validate_avatar_url(avatar_url)
        
        # discriminator
        if discriminator is ...:
            discriminator = 0
        else:
            discriminator = validate_discriminator(discriminator)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # status
        if status is ...:
            status = Status.offline
        else:
            status = validate_status(status)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self.activity_name = activity_name
        self.avatar_url = avatar_url
        self.discriminator = discriminator
        self.id = user_id
        self.name = name
        self.status = status
        return self
        
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild widget user from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Guild widget user data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.activity_name = parse_activity_name(data)
        self.avatar_url = parse_avatar_url(data)
        self.discriminator = parse_discriminator(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.status = parse_status(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the guild widget user.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_activity_name_into(self.activity_name, data, defaults)
        put_avatar_url_into(self.avatar_url, data, defaults)
        put_discriminator_into(self.discriminator, data, defaults)
        put_id_into(self.id, data, defaults)
        put_name_into(self.name, data, defaults)
        put_status_into(self.status, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the representation of the guild widget user."""
        return f'<{self.__class__.__name__} id = {self.id}, name = {self.full_name!r}>'
    
    
    def __hash__(self):
        """Returns the guild widget user's hash value."""
        hash_value = 0
        
        # activity_name
        activity_name = self.activity_name
        if (activity_name is not None):
            hash_value ^= hash(activity_name)
        
        # avatar_url
        hash_value ^= hash(self.avatar_url)
        
        # discriminator
        hash_value ^= self.discriminator
        
        # id
        hash_value ^= self.id << 6
        
        # name
        hash_value ^= hash(self.name)
        
        # status
        hash_value ^= hash(self.status)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two guild widget users are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two guild widget users are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # activity_name
        if self.activity_name != other.activity_name:
            return False
        
        # avatar_url
        if self.avatar_url != other.avatar_url:
            return False
        
        # discriminator
        if self.discriminator != other.discriminator:
            return False
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # status
        if self.status is not other.status:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the guild widget user.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.activity_name = self.activity_name
        new.avatar_url = self.avatar_url
        new.discriminator = self.discriminator
        new.id = self.id
        new.name = self.name
        new.status = self.status
        return new
    
    
    def copy_with(
        self,
        *,
        activity_name = ...,
        avatar_url = ...,
        discriminator = ...,
        name = ...,
        status = ...,
        user_id = ...,
    ):
        """
        Copies the guild widget user with the given fields.
        
        Parameters
        ----------
        activity_name : `None`, `str`, Optional (Keyword only)
            The guild widget user's activity's name if applicable.
        avatar_url : `None`, `str`, Optional (Keyword only)
            The guild widget user's avatar url.
        discriminator : `int`, Optional (Keyword only)
            The guild widget user's discriminator.
        name : `str`, Optional (Keyword only)
            The guild widget user's name.
        status : ``Status``, `str`, Optional (Keyword only)
            The guild widget user's status.
        user_id : `int`, Optional (Keyword only)
            The unique identifier number of the guild widget user. Can be between `0` and `99`.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # activity_name
        if activity_name is ...:
            activity_name = self.activity_name
        else:
            activity_name = validate_activity_name(activity_name)
        
        # avatar_url
        if avatar_url is ...:
            avatar_url = self.avatar_url
        else:
            avatar_url = validate_avatar_url(avatar_url)
        
        # discriminator
        if discriminator is ...:
            discriminator = self.discriminator
        else:
            discriminator = validate_discriminator(discriminator)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # status
        if status is ...:
            status = self.status
        else:
            status = validate_status(status)
        
        # user_id
        if user_id is ...:
            user_id = self.id
        else:
            user_id = validate_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new.activity_name = activity_name
        new.avatar_url = avatar_url
        new.discriminator = discriminator
        new.id = user_id
        new.name = name
        new.status = status
        return new
        
        
    @property
    def full_name(self):
        """
        The user's name with it's discriminator (if applicable).
        
        Returns
        -------
        full_name : `str`
        """
        discriminator = self.discriminator
        if discriminator:
            return f'{self.name}#{discriminator:0>4}'
        
        return self.name
    
    
    @property
    def mention(self):
        """
        The mention of the user.
        
        Returns
        -------
        mention : `str`
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.mention` is deprecated and will be removed in 2023 November. '
                f'This property actually never returned the correct value, that is why.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return f'<@{self.id}>'
    
    
    @property
    def mention_nick(self):
        """
        The mention to the user's nick.
        
        Returns
        -------
        mention : `str`
        
        Notes
        -----
        It actually has nothing to do with the user's nickname > <.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.mention_nick` is deprecated and will be removed in 2023 November. '
                f'This property actually never returned the correct value, that is why.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return f'<@!{self.id}>'



__all__ = ()

from ...field_parsers import entity_id_array_parser_factory, nullable_date_time_parser_factory
from ...field_putters import nullable_date_time_optional_putter_factory, nullable_entity_id_array_putter_factory
from ...field_validators import entity_id_array_validator_factory, nullable_date_time_validator_factory
from ...user import ClientUserBase

# ended_at

parse_ended_at = nullable_date_time_parser_factory('ended_timestamp')
put_ended_at_into = nullable_date_time_optional_putter_factory('ended_timestamp')
validate_ended_at = nullable_date_time_validator_factory('ended_at')

# user_ids

parse_user_ids = entity_id_array_parser_factory('participants')
put_user_ids_into = nullable_entity_id_array_putter_factory('participants')
validate_user_ids = entity_id_array_validator_factory('user_ids', ClientUserBase)



__all__ = ('ApplicationCommandPermissionOverwrite',)

from scarletio import RichAttributeErrorBaseType

from ...channel import ChannelType, create_partial_channel_from_id
from ...core import CHANNELS, ROLES
from ...role import create_partial_role_from_id
from ...user import create_partial_user_from_id

from .fields import (
    parse_allow, parse_target_id, parse_target_type, put_allow_into, put_target_id_into, put_target_type_into,
    validate_allow
)
from .helpers import validate_application_command_permission_overwrite_target
from .preinstanced import ApplicationCommandPermissionOverwriteTargetType


APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER = ApplicationCommandPermissionOverwriteTargetType.user
APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE = ApplicationCommandPermissionOverwriteTargetType.role
APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL = ApplicationCommandPermissionOverwriteTargetType.channel


class ApplicationCommandPermissionOverwrite(RichAttributeErrorBaseType):
    """
    Represents an application command's allow/disallow overwrite for the given entity.
    
    Attributes
    ----------
    allow : `bool`
        Whether the respective command is allowed for the represented entity.
    
    target_id : `int`
        The represented entity's identifier.
    
    target_type : ``ApplicationCommandPermissionOverwriteTargetType``
        The target entity's type.
    """
    __slots__ = ('allow', 'target_id', 'target_type')
    
    def __new__(cls, target, allow):
        """
        Creates a new application command permission overwrite with the given parameters.
        
        Parameters
        ----------
        target : ``ClientUserBase``, ``Role``, ``Channel``, `tuple` ((``ClientUserBase``, ``Role``, \
                ``Channel``, `str` (`'Role'`, `'role'`, `'User'`, `'user'`, `'Channel'`, `'channel'`, \
                ``ApplicationCommandPermissionOverwriteTargetType``, `int`)), `int`)
            The target entity of the application command permission overwrite.
            
            The expected type & value might be pretty confusing, but the target was it to allow relaxing creation.
            To avoid confusing, here is a list of the expected structures:
            
            - ``Role``
            - ``ClientUserBase``
            - ``Channel``
            - `tuple` (``Role``, `int`)
            - `tuple` (``ClientUserBase``, `int`)
            - `tuple` (``Channel``, `int`)
            - `tuple` (`'Role'`, `int`)
            - `tuple` (`'role'`, `int`)
            - `tuple` (`'User'`, `int`)
            - `tuple` (`'user'`, `int`)
            - `tuple` (`'Channel'`, `int`)
            - `tuple` (`'channel'`, `int`)
            - `tuple` (``ApplicationCommandPermissionOverwriteTargetType``, `int`)
            - `tuple` (`int`, `int`)
        
        allow : `bool`
            Whether the respective application command should be enabled for the respective entity.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        AssertionError
            - If a parameter's value is incorrect.
        """
        # allow
        allow = validate_allow(allow)
        
        # target_id & target_type
        target_type, target_id = validate_application_command_permission_overwrite_target(target)
        
        self = object.__new__(cls)
        self.allow = allow
        self.target_id = target_id
        self.target_type = target_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandPermissionOverwrite`` from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            The received application command permission overwrite data.
        
        Returns
        -------
        self : `instance<cls>`
            The created application command option.
        """
        self = object.__new__(cls)
        self.allow = parse_allow(data)
        self.target_id = parse_target_id(data)
        self.target_type = parse_target_type(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application command permission overwrite to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_allow_into(self.allow, data, defaults)
        put_target_id_into(self.target_id, data, defaults)
        put_target_type_into(self.target_type, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the application command permission overwrite's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Primary fields: `.target_type`, `.target_id`
        repr_parts.append(' target_type = ')
        repr_parts.append(self.target_type.name)
        
        repr_parts.append(', target_id = ')
        repr_parts.append(repr(self.target_id))
        
        # Secondary fields: `.allow`
        repr_parts.append(', allow = ')
        repr_parts.append(repr(self.allow))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application command overwrites are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # allow
        if self.allow != other.allow:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        # target_type
        if self.target_type is not other.target_type:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the application command permission overwrite's hash value."""
        hash_value = 0
        
        # allow
        hash_value ^= self.allow << 8
        
        # target_id
        hash_value ^= self.target_id
        
        # target_type
        hash_value ^= self.target_type.value
        
        return hash_value
    
    
    def __gt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_type_value = self.target_type.value
        other_type_value = other.target_type.value
        
        if self_type_value > other_type_value:
            return True
        
        if self_type_value < other_type_value:
            return False
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_role = ROLES.get(self_target_id, None)
            other_role = ROLES.get(other_target_id, None)
            if self_role is None:
                if other_role is None:
                    return (self_target_id > other_target_id)
                else:
                    return False
            else:
                if other_role is None:
                    return True
                else:
                    return (self_role > other_role)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER.value:
            return (self.target_id > other.target_id)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_channel = CHANNELS.get(self_target_id, None)
            other_channel = CHANNELS.get(other_target_id, None)
            if self_channel is None:
                if other_channel is None:
                    return (self_target_id > other_target_id)
                else:
                    return False
            else:
                if other_channel is None:
                    return True
                else:
                    return (self_channel.name > other_channel.name)
        
        # Should not happen
        return False
    
    
    def __lt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_type_value = self.target_type.value
        other_type_value = other.target_type.value
        
        if self_type_value > other_type_value:
            return False
        
        if self_type_value < other_type_value:
            return True
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_role = ROLES.get(self_target_id, None)
            other_role = ROLES.get(other_target_id, None)
            if self_role is None:
                if other_role is None:
                    return (self_target_id < other_target_id)
                else:
                    return True
            else:
                if other_role is None:
                    return False
                else:
                    return (self_role < other_role)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER.value:
            return (self.target_id < other.target_id)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL.value:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_channel = CHANNELS.get(self_target_id, None)
            other_channel = CHANNELS.get(other_target_id, None)
            if self_channel is None:
                if other_channel is None:
                    return (self_target_id < other_target_id)
                else:
                    return True
            else:
                if other_channel is None:
                    return False
                else:
                    return (self_channel.name < other_channel.name)
        
        # Should not happen
        return True
    
    
    def copy(self):
        """
        Copies the application command permission overwrite.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # allow
        new.allow = self.allow
        
        # target_id
        new.target_id = self.target_id
        
        # target_type
        new.target_type = self.target_type
        
        return new
    
    
    
    def copy_with(self, *, allow = ..., target = ...,):
        """
        Copies the application command permission overwrite with modifying the given parameters inside of it.
        
        Parameters
        ----------
        allow : `bool`
            Whether the respective application command should be enabled for the respective entity.
        
        target : ``ClientUserBase``, ``Role``, ``Channel``, `tuple` ((``ClientUserBase``, ``Role``, \
                ``Channel``, `str` (`'Role'`, `'role'`, `'User'`, `'user'`, `'Channel'`, `'channel'`, \
                ``ApplicationCommandPermissionOverwriteTargetType``, `int`)), `int`)
            The target entity of the application command permission overwrite.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        AssertionError
            - If a parameter's value is incorrect.
        """
        # allow
        if allow is ...:
            allow = self.allow
        else:
            allow = validate_allow(allow)
        
        if target is ...:
            target_id = self.target_id
            target_type = self.target_type
        else:
            target_type, target_id = validate_application_command_permission_overwrite_target(target)
        
        new = object.__new__(type(self))
        new.allow = allow
        new.target_id = target_id
        new.target_type = target_type
        return new
    
    
    @property
    def target(self):
        """
        Returns the application command overwrite's target entity.
        
        Returns
        -------
        target : `None`, ``Role``, ``ClientUserBase``, ``Channel``
        """
        target_type = self.target_type
        target_id = self.target_id
        
        if target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE:
            target = create_partial_role_from_id(target_id)
        
        elif target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER:
            target = create_partial_user_from_id(target_id)
        
        elif target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_CHANNEL:
            target = create_partial_channel_from_id(target_id, ChannelType.unknown, 0)
        
        else:
            target = None
        
        return target



from .command_base_application_command import *
from .constants import *
from .helpers import *

__all__ = (
    *command_base_application_command.__all__,
    *constants.__all__,
    *helpers.__all__,
)



__all__ = ('RoleManagerMetadataBooster',)

from scarletio import copy_docs

from .base import RoleManagerMetadataBase
from .constants import BOOSTER_KEY


class RoleManagerMetadataBooster(RoleManagerMetadataBase):
    """
    Role manager metadata of a booster role.
    """
    __slots__ = ()
    
    @copy_docs(RoleManagerMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        return {BOOSTER_KEY: None}



__all__ = ('GuildUserChunkEvent',)

import warnings

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS

from .fields import (
    parse_chunk_count, parse_chunk_index, parse_guild_id, parse_nonce, parse_users, put_chunk_count_into,
    put_chunk_index_into, put_guild_id_into, put_nonce_into, put_users_into, validate_chunk_count, validate_chunk_index,
    validate_guild_id, validate_nonce, validate_users
)


class GuildUserChunkEvent(EventBase):
    """
    Represents a processed `GUILD_MEMBERS_CHUNK` dispatch event.
    
    Attributes
    ----------
    chunk_count : `int`
        The total number of chunk responses what Discord sends for the respective gateway.
    chunk_index : `int`
        The chunk_index of the received chunk response (0 <= chunk_index < chunk_count).
    guild_id : `int`
        The guild's identifier, what received the user chunk.
    nonce : `None`, `str`
        A nonce to identify guild user chunk response.
    users : `list` of ``ClientUserBase``
        The received users.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, users, nonce, chunk_index, chunk_count = event
    ```
    """
    __slots__ = ('chunk_count', 'chunk_index', 'guild_id', 'nonce', 'users')
    
    def __new__(cls, *, chunk_count = ..., chunk_index = ..., guild_id = ..., nonce = ..., users = ...):
        """
        Creates a new guild chunk event from the given parameters.
        
        Parameters
        ----------
        chunk_count : `int`, Optional (Keyword only)
            The total number of chunk responses what Discord sends for the respective gateway.
        chunk_index : `int`, Optional (Keyword only)
            The chunk_index of the received chunk response (0 <= chunk_index < chunk_count).
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier, what received the user chunk.
        nonce : `None`, `str`, Optional (Keyword only)
            A nonce to identify guild user chunk response.
        users : `None`, `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The received users.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # chunk_count
        if chunk_count is ...:
            chunk_count = 0
        else:
            chunk_count = validate_chunk_count(chunk_count)
        
        # chunk_index
        if chunk_index is ...:
            chunk_index = 0
        else:
            chunk_index = validate_chunk_index(chunk_index)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # nonce
        if nonce is ...:
            nonce = None
        else:
            nonce = validate_nonce(nonce)
        
        # users
        if users is ...:
            users = []
        else:
            users = validate_users(users)
        
        self = object.__new__(cls)
        self.chunk_count = chunk_count
        self.chunk_index = chunk_index
        self.guild_id = guild_id
        self.nonce = nonce
        self.users = users
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild user chunk event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Guild user chunk event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(GuildUserChunkEvent)
        self.chunk_index = parse_chunk_index(data)
        self.chunk_count = parse_chunk_count(data)
        self.guild_id = guild_id = parse_guild_id(data)
        self.nonce = parse_nonce(data)
        self.users = parse_users(data, guild_id)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the guild user chunk event into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_chunk_count_into(self.chunk_count, data, defaults)
        put_chunk_index_into(self.chunk_index, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        put_nonce_into(self.nonce, data, defaults)
        put_users_into(self.users, data, defaults, guild_id = self.guild_id)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        chunk_count = self.chunk_count
        if chunk_count != 1:
            repr_parts.append(', chunk_index = ')
            repr_parts.append(repr(self.chunk_index))
            repr_parts.append('/')
            repr_parts.append(repr(chunk_count))
        
        nonce = self.nonce
        if (nonce is not None):
            repr_parts.append(' nonce = ')
            repr_parts.append(repr(nonce))
        
        repr_parts.append(', user_count: ')
        repr_parts.append(repr(len(self.users)))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 5
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.users
        yield self.nonce
        yield self.chunk_index
        yield self.chunk_count
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.chunk_count != other.chunk_count:
            return False
        
        if self.chunk_index != other.chunk_index:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.nonce != other.nonce:
            return False
        
        if self.users != other.users:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # chunk_count
        hash_value ^= self.chunk_count
        
        # chunk_index
        hash_value ^= (self.chunk_index << 12)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # nonce
        nonce = self.nonce
        if (nonce is not None):
            hash_value ^= hash(nonce)
        
        # users
        users = self.users
        if users:
            hash_value ^= (len(users) << 24)
            
            for user in users:
                hash_value ^= user.id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the guild user chunk event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.chunk_count = self.chunk_count
        new.chunk_index = self.chunk_index
        new.guild_id = self.guild_id
        new.nonce = self.nonce
        new.users = [*self.users]
        return new
    
    
    def copy_with(self, *, chunk_count = ..., chunk_index = ..., guild_id = ..., nonce = ..., users = ...):
        """
        Copies the guild chunk event with the given fields.
        
        Parameters
        ----------
        chunk_count : `int`, Optional (Keyword only)
            The total number of chunk responses what Discord sends for the respective gateway.
        chunk_index : `int`, Optional (Keyword only)
            The chunk_index of the received chunk response (0 <= chunk_index < chunk_count).
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier, what received the user chunk.
        nonce : `None`, `str`, Optional (Keyword only)
            A nonce to identify guild user chunk response.
        users : `None`, `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The received users.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # chunk_count
        if chunk_count is ...:
            chunk_count = self.chunk_count
        else:
            chunk_count = validate_chunk_count(chunk_count)
        
        # chunk_index
        if chunk_index is ...:
            chunk_index = self.chunk_index
        else:
            chunk_index = validate_chunk_index(chunk_index)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # nonce
        if nonce is ...:
            nonce = self.nonce
        else:
            nonce = validate_nonce(nonce)
        
        # users
        if users is ...:
            users = [*self.users]
        else:
            users = validate_users(users)
        
        new = object.__new__(type(self))
        new.chunk_count = chunk_count
        new.chunk_index = chunk_index
        new.guild_id = guild_id
        new.nonce = nonce
        new.users = users
        return new
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    def iter_users(self):
        """
        Iterates over the received users.
        
        This method is a generator.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        yield from self.users
    
    
    @property
    def index(self):
        warnings.warn(
            (
                f'`{self.__name__}.index` is deprecated and will be removed in 2023 December.'
                f'Please use `.chunk_index` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.chunk_index
    
    
    @property
    def count(self):
        warnings.warn(
            (
                f'`{self.__name__}.count` is deprecated and will be removed in 2023 December.'
                f'Please use `.chunk_count` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.chunk_count



__all__ = ()

from ...field_parsers import force_string_parser_factory, int_parser_factory
from ...field_putters import force_string_putter_factory, int_putter_factory
from ...field_validators import (
    force_date_time_validator_factory, force_string_validator_factory, int_conditional_validator_factory,
    preinstanced_array_validator_factory, url_required_validator_factory
)

from .preinstanced import Oauth2Scope

# access_token

parse_access_token = force_string_parser_factory('access_token')
put_access_token_into = force_string_putter_factory('access_token')
validate_access_token = force_string_validator_factory('access_token', 0, 1024)

# created_at

validate_created_at = force_date_time_validator_factory('created_at')

# expires_after

parse_expires_after = int_parser_factory('expires_in', 0)
put_expires_after_into = int_putter_factory('expires_in')
validate_expires_after = int_conditional_validator_factory(
    'expires_after',
    0,
    lambda expires_after : expires_after >= 0,
    '>= 0',
)

# refresh_token

parse_refresh_token = force_string_parser_factory('refresh_token')
put_refresh_token_into = force_string_putter_factory('refresh_token')
validate_refresh_token = force_string_validator_factory('refresh_token', 0, 1024)

# redirect_url

validate_redirect_url = url_required_validator_factory('redirect_url')

# scopes

def parse_scopes(data):
    """
    Parses the oauth2 scopes from the given data.
    
    Parameters
    ----------
    joined_scopes : `str`
        The joined scopes.
    
    Returns
    -------
    scopes : `None`, `tuple` of ``Oauth2Scope``
    """
    joined_scopes = data.get('scope', None)
    if joined_scopes is None:
        return None
    
    split_scopes = joined_scopes.split()
    if not split_scopes:
        return None
    
    split_scopes.sort()
    
    return tuple(Oauth2Scope.get(scope) for scope in split_scopes)


def put_scopes_into(scopes, data, defaults):
    """
    Puts the given oauth2 scopes into a json serializable dictionary.
    
    Parameters
    ----------
    activity_id : `int`
        Activity's identifier.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if scopes is None:
        joined_scopes = ''
    else:
        joined_scopes = ' '.join([scope.value for scope in scopes])
        
    data['scope'] = joined_scopes
    return data

validate_scopes = preinstanced_array_validator_factory('scopes', Oauth2Scope)



__all__ = ()

from functools import partial as partial_func

from ...field_parsers import (
    force_string_parser_factory, nullable_functional_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    force_string_putter_factory, nullable_functional_optional_putter_factory, nullable_string_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    force_string_validator_factory, nullable_string_validator_factory, preinstanced_validator_factory
)
from ...localization.helpers import localized_dictionary_builder
from ...localization.utils import build_locale_dictionary, destroy_locale_dictionary

from .constants import (
    DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN, KEY_LENGTH_MAX, KEY_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN
)
from .preinstanced import ApplicationRoleConnectionMetadataType

# description 

parse_description = nullable_string_parser_factory('description')
put_description_into = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX)

# description_localizations

parse_description_localizations = nullable_functional_parser_factory(
    'description_localizations', build_locale_dictionary
)
put_description_localizations_into = nullable_functional_optional_putter_factory(
    'description_localizations', destroy_locale_dictionary
)
validate_description_localizations = partial_func(
    localized_dictionary_builder, parameter_name = 'description_localizations'
)

# key 

parse_key = force_string_parser_factory('key')
put_key_into = force_string_putter_factory('key')
validate_key = force_string_validator_factory('key', KEY_LENGTH_MIN, KEY_LENGTH_MAX)

# name 

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# name_localizations

parse_name_localizations = nullable_functional_parser_factory(
    'name_localizations', build_locale_dictionary
)
put_name_localizations_into = nullable_functional_optional_putter_factory(
    'name_localizations', destroy_locale_dictionary
)
validate_name_localizations = partial_func(
    localized_dictionary_builder, parameter_name = 'name_localizations'
)

# type

parse_type = preinstanced_parser_factory(
    'type', ApplicationRoleConnectionMetadataType, ApplicationRoleConnectionMetadataType.none
)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', ApplicationRoleConnectionMetadataType)



__all__ = ('OnboardingScreen',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_default_channel_ids, parse_enabled, parse_guild_id, parse_prompts, put_default_channel_ids_into,
    put_enabled_into, put_guild_id_into, put_prompts_into, validate_default_channel_ids, validate_enabled,
    validate_guild_id, validate_prompts
)
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...core import GUILDS
from ...channel import ChannelType, create_partial_channel_from_id


PRECREATE_FIELDS = {
    'default_channel_ids': ('default_channel_ids', validate_default_channel_ids),
    'default_channels': ('default_channel_ids', validate_default_channel_ids),
    'enabled': ('enabled', validate_enabled),
    'guild_id': ('guild_id', validate_guild_id),
    'guild': ('guild_id', validate_guild_id),
    'prompts': ('prompts', validate_prompts),
}


class OnboardingScreen(RichAttributeErrorBaseType):
    """
    Represents a guild onboarding screen.
    
    Attributes
    ----------
    default_channel_ids : `None`, `tuple` of `int`
        The channels' identifiers that new members get opted into automatically.
    enabled : `bool`
        Whether onboarding is enabled.
    guild_id : `int`
        The guild's identifier the onboarding screen is part of.
    prompts : `None`, `tuple` of ``OnboardingPrompt``
        The prompts shown during onboarding and in customize community.
    """
    __slots__ = ('default_channel_ids', 'enabled', 'guild_id', 'prompts',)
    
    def __new__(cls, *, default_channel_ids = ..., enabled = ..., prompts = ...):
        """
        Creates an onboarding screen instance from the given parameters.
        
        Parameters
        ----------
        default_channel_ids : `None`, `iterable` of (`int`, ``Channel``), Optional (Keyword only)
            The channels' identifiers that new members get opted into automatically.
        enabled : `bool`, Optional (Keyword only)
            Whether onboarding is enabled.
        prompts : `None`, `iterable` of ``OnboardingPrompt``, Optional (Keyword only)
            The prompts shown during onboarding and in customize community.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default_channel_ids
        if default_channel_ids is ...:
            default_channel_ids = None
        else:
            default_channel_ids = validate_default_channel_ids(default_channel_ids)
        
        # enabled
        if enabled is ...:
            enabled = True
        else:
            enabled = validate_enabled(enabled)
        
        # prompts
        if prompts is ...:
            prompts = None
        else:
            prompts = validate_prompts(prompts)
        
        self = object.__new__(cls)
        self.default_channel_ids = default_channel_ids
        self.enabled = enabled
        self.guild_id = 0
        self.prompts = prompts
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new onboarding screen instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Auto moderation rule trigger metadata payload.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.default_channel_ids = parse_default_channel_ids(data)
        self.enabled = parse_enabled(data)
        self.guild_id = parse_guild_id(data)
        self.prompts = parse_prompts(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the onboarding screen to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default fields should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields (like id-s) should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_default_channel_ids_into(self.default_channel_ids, data, defaults)
        put_enabled_into(self.enabled, data, defaults)
        put_prompts_into(self.prompts, data, defaults, include_internals = include_internals)
        
        if include_internals:
            put_guild_id_into(self.guild_id, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the onboarding screen's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        guild_id = self.guild_id
        if guild_id:
            repr_parts.append(' guild_id = ')
            repr_parts.append(repr(guild_id))
            repr_parts.append(',')
        
        repr_parts.append(' enabled = ')
        repr_parts.append(repr(self.enabled))
        
        default_channel_ids = self.default_channel_ids
        if (default_channel_ids is not None):
            repr_parts.append(', default_channel_ids = [')
            
            index = 0
            length = len(default_channel_ids)
            
            while True:
                channel_id = default_channel_ids[index]
                index += 1
                
                repr_parts.append(repr(channel_id))
                
                if index == length:
                    break
                
                repr_parts.append(', ')
            
            repr_parts.append(']')
        
        prompts = self.prompts
        if (prompts is not None):
            repr_parts.append(', prompts = [')
            
            index = 0
            length = len(prompts)
            
            while True:
                prompt = prompts[index]
                index += 1
                
                repr_parts.append(repr(prompt))
                
                if index == length:
                    break
                
                repr_parts.append(', ')
            
            repr_parts.append(']')
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two onboarding screen's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.default_channel_ids != other.default_channel_ids:
            return False
        
        if self.enabled != other.enabled:
            return False
        
        # guild_id
        # ignore, internal
        
        if self.prompts != other.prompts:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the onboarding screen's hash value ."""
        hash_value = 0
        
        default_channel_ids = self.default_channel_ids
        if (default_channel_ids is not None):
            hash_value ^= len(default_channel_ids)
            
            for channel_id in default_channel_ids:
                hash_value ^= channel_id
        
        hash_value ^= self.enabled << 4
        
        # guild_id
        # Ignore it
        
        prompts = self.prompts
        if (prompts is not None):
            hash_value ^= len(prompts) << 5
            
            hash_value ^= hash(prompts)
            hash_value ^= (1 << 12)
        
        return hash_value
    
    
    @classmethod
    def precreate(cls, **keyword_parameters):
        """
        Precreates an onboarding screen. Since they are not cached, this method just a ``.__new__`` alternative.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Additional parameters defining how the option's fields should be set.
        
        Other Parameters
        ----------------
        default_channel_ids : `None`, `iterable` of (`int`, ``Channel``), Optional (Keyword only)
            The channels' identifiers that new members get opted into automatically.
        default_channels : `None`, `iterable` of (`int`, ``Channel``), Optional (Keyword only)
            Alternative of `default_channel_ids`.
        enabled : `bool`, Optional (Keyword only)
            Whether onboarding is enabled.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier the onboarding screen is part of.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            Alternative of `guild_id`.
        prompts : `None`, `iterable` of ``OnboardingPrompt``, Optional (Keyword only)
            The prompts shown during onboarding and in customize community.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        # Construct
        
        self = object.__new__(cls)
        self.default_channel_ids = None
        self.enabled = True
        self.guild_id = 0
        self.prompts = None
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def copy(self):
        """
        Copies the onboarding screen.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        
        default_channel_ids = self.default_channel_ids
        if (default_channel_ids is not None):
            default_channel_ids = (*default_channel_ids,)
        new.default_channel_ids = default_channel_ids
        
        new.enabled = self.enabled
        new.guild_id = 0
        
        prompts = self.prompts
        if (prompts is not None):
            prompts = (*(prompt.copy() for prompt in prompts),)
        new.prompts = prompts
        
        return new
    
    
    def copy_with(self, *, default_channel_ids = ..., enabled = ..., prompts = ...):
        """
        Copies the onboarding screen with the given screen.
        
        Parameters
        ----------
        Parameters
        ----------
        default_channel_ids : `None`, `iterable` of (`int`, ``Channel``), Optional (Keyword only)
            The channels' identifiers that new members get opted into automatically.
        enabled : `bool`, Optional (Keyword only)
            Whether onboarding is enabled.
        prompts : `None`, `iterable` of ``OnboardingPrompt``, Optional (Keyword only)
            The prompts shown during onboarding and in customize community.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default_channel_ids
        if default_channel_ids is ...:
            default_channel_ids = self.default_channel_ids
            if (default_channel_ids is not None):
                default_channel_ids = (*default_channel_ids,)
        else:
            default_channel_ids = validate_default_channel_ids(default_channel_ids)
        
        # enabled
        if enabled is ...:
            enabled = self.enabled
        else:
            enabled = validate_enabled(enabled)
        
        # prompts
        if prompts is ...:
            prompts = self.prompts
            if (prompts is not None):
                prompts = (*(prompt.copy() for prompt in prompts),)
        else:
            prompts = validate_prompts(prompts)
        
        new = object.__new__(type(self))
        new.default_channel_ids = default_channel_ids
        new.enabled = enabled
        new.guild_id = 0
        new.prompts = prompts
        return new
    
    
    @property
    def default_channels(self):
        """
        Returns the default channels that new get opted into automatically.
        
        Returns
        -------
        default_channels : `None`, `tuple` of ``Channel``
        """
        default_channel_ids = self.default_channel_ids
        if default_channel_ids is None:
            default_channels = None
        else:
            default_channels = (*(
                create_partial_channel_from_id(channel_id, ChannelType.unknown, 0) for channel_id in default_channel_ids
            ),)
        
        return default_channels
    
    
    def iter_default_channel_ids(self):
        """
        Iterates over the default channels that new get opted into automatically.
        
        This method is an iterable generator.
        
        Yields
        ------
        default_channel_id : `int`
        """
        default_channel_ids = self.default_channel_ids
        if (default_channel_ids is not None):
            yield from default_channel_ids
    
    
    def iter_default_channels(self):
        """
        Iterates over the default channels that new get opted into automatically.
        
        This method is an iterable generator.
        
        Yields
        ------
        default_channel : ``Channel``
        """
        for channel_id in self.iter_default_channel_ids():
            yield create_partial_channel_from_id(channel_id, ChannelType.unknown, 0)
    
    
    @property
    def guild(self):
        """
        Returns the onboarding screen's owner guild identifier. If the guild is not cached then returns `None`.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS[guild_id]
    
    
    def iter_prompts(self):
        """
        Iterates over the prompts that are shown during onboarding and in customize community.
        
        This method is an iterable generator.
        
        Yields
        ------
        prompt : ``OnboardingPrompt``
        """
        prompts = self.prompts
        if (prompts is not None):
            yield from prompts



__all__ = ()

from .fields import (
    put_privacy_level_into, put_send_start_notification_into, put_topic_into, validate_privacy_level,
    validate_send_start_notification, validate_topic
)


STAGE_CREATE_FIELD_CONVERTERS = {
    'privacy_level': (validate_privacy_level, put_privacy_level_into),
    'send_start_notification': (validate_send_start_notification, put_send_start_notification_into),
    'topic': (validate_topic, put_topic_into),
}

STAGE_EDIT_FIELD_CONVERTERS = {
    'privacy_level': (validate_privacy_level, put_privacy_level_into),
    'topic': (validate_topic, put_topic_into),
}



__all__ = ()

from ...field_parsers import bool_parser_factory, entity_id_parser_factory, preinstanced_parser_factory
from ...field_putters import entity_id_putter_factory, force_bool_putter_factory, preinstanced_putter_factory
from ...field_validators import bool_validator_factory, entity_id_validator_factory, preinstanced_validator_factory

from .preinstanced import ApplicationCommandPermissionOverwriteTargetType


# allow

parse_allow = bool_parser_factory('permission', True)
put_allow_into = force_bool_putter_factory('permission')
validate_allow = bool_validator_factory('allow', True)

# target_id

parse_target_id = entity_id_parser_factory('id')
put_target_id_into = entity_id_putter_factory('id')
validate_target_id = entity_id_validator_factory('target_id')

# target_type

parse_target_type = preinstanced_parser_factory(
    'type', ApplicationCommandPermissionOverwriteTargetType, ApplicationCommandPermissionOverwriteTargetType.none
)
put_target_type_into = preinstanced_putter_factory('type')
validate_target_type = preinstanced_validator_factory('target_type', ApplicationCommandPermissionOverwriteTargetType)



from .interaction_component import *
from .interaction_event import *
from .interaction_metadata import *
from .interaction_option import *
from .resolved import *
from .responding import *


__all__ = (
    *interaction_component.__all__,
    *interaction_event.__all__,
    *interaction_metadata.__all__,
    *interaction_option.__all__,
    *resolved.__all__,
    *responding.__all__,
)



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_array_parser_factory, flag_parser_factory, nullable_date_time_parser_factory,
    nullable_string_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, flag_optional_putter_factory, nullable_date_time_optional_putter_factory,
    optional_entity_id_array_optional_putter_factory, nullable_string_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, flag_validator_factory, entity_id_array_validator_factory,
    nullable_date_time_validator_factory, nullable_string_validator_factory
)

from .constants import NICK_LENGTH_MAX, NICK_LENGTH_MIN
from .flags import GuildProfileFlag

# boosts_since

parse_boosts_since = nullable_date_time_parser_factory('premium_since')
put_boosts_since_into = nullable_date_time_optional_putter_factory('premium_since')
validate_boosts_since = nullable_date_time_validator_factory('boosts_since')

# flags

parse_flags = flag_parser_factory('flags', GuildProfileFlag)
put_flags_into = flag_optional_putter_factory('flags', GuildProfileFlag())
validate_flags = flag_validator_factory('flags', GuildProfileFlag)

# joined_at

parse_joined_at = nullable_date_time_parser_factory('joined_at')
put_joined_at_into = nullable_date_time_optional_putter_factory('joined_at')
validate_joined_at = nullable_date_time_validator_factory('joined_at')

# nick

parse_nick = nullable_string_parser_factory('nick')
put_nick_into = nullable_string_optional_putter_factory('nick')
validate_nick = nullable_string_validator_factory('nick', NICK_LENGTH_MIN, NICK_LENGTH_MAX)

# pending

parse_pending = bool_parser_factory('pending', False)
put_pending_into = bool_optional_putter_factory('pending', False)
validate_pending = bool_validator_factory('pending', False)

# role_ids

parse_role_ids = entity_id_array_parser_factory('roles')
put_role_ids_into = optional_entity_id_array_optional_putter_factory('roles')
validate_role_ids = entity_id_array_validator_factory('role_ids', NotImplemented, include = 'Role')

# timed_out_until

parse_timed_out_until = nullable_date_time_parser_factory('communication_disabled_until')
put_timed_out_until_into = nullable_date_time_optional_putter_factory('communication_disabled_until')
validate_timed_out_until = nullable_date_time_validator_factory('timed_out_until')



__all__ = ()

import sys
from os import listdir as list_directory, sep as PATH_SEPARATOR
from os.path import (
    basename as base_name, dirname as get_parent_directory_path, exists, isabs as is_absolute_path,
    isdir as is_directory, isfile as is_file, join as join_paths
)

from scarletio import CallableAnalyzer, HybridValueDictionary, include

from .constants import ABSOLUTE_PATH_PLUGIN_NAME_PREFIX, PLUGIN_ROOTS
from .plugin_root import is_tuple_starting_with, register_plugin_root


PLUGIN_LOADER = include('PLUGIN_LOADER')


def _validate_entry_or_exit(point):
    """
    Validates the given entry or exit point, returning `True`, if they passed.
    
    Parameters
    ----------
    point : `None`, `str`, `callable`
        The point to validate.
    
    Raises
    ------
    TypeError
        If `point` was given as `callable`, but accepts less or more positional parameters, as would be given.
    """
    if point is None:
        return True
    
    if isinstance(point, str):
        return True
    
    if callable(point):
        analyzer = CallableAnalyzer(point)
        min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
        if min_ > 1:
            raise TypeError(
                f'`{point!r}` excepts at least `{min_!r}` non reserved parameters, meanwhile the event '
                f'expects to pass `1`, got {point!r}'
            )
        
        if min_ == 1:
            return True
        
        #min<expected
        if max_ >= 1:
            return True
        
        if analyzer.accepts_args():
            return True
        
        raise TypeError(
            f'`{point!r}` expects maximum `{max_!r}` non reserved parameters  meanwhile the event expects '
            f'to pass `1`., got {point!r}'
        )
    
    return False


def validate_plugin_parameters(
    entry_point = None,
    exit_point = None,
    extend_default_variables = True,
    locked = False,
    take_snapshot_difference = True,
    **variables,
):
    """
    Validates plugin parameters.
    
    Parameters
    ----------
    entry_point : `None`, `str`, `callable`, = `None` Optional
        Plugin specific entry point, to use over the plugin loader's default.
    exit_point : `None`, `str`, `callable` = `None`, Optional
        Plugin specific exit point, to use over the plugin loader's default.
    extend_default_variables : `bool` = `True`, Optional
        Whether the plugin should use the loader's default variables or just it's own.
    locked : `bool` = `False`, Optional
        Whether the given plugin(s) should not be affected by `.{}_all` methods.
    take_snapshot_difference : `bool` = `True`, Optional
        Whether snapshot feature should be used.
    **variables : Keyword parameters
        Variables to assign to an plugin(s)'s module before they are loaded.
    
    Raises
    ------
    TypeError
        - If `entry_point` was not given as `None`, `str`, `callable`.
        - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
            given.
        - If `exit_point` was not given as `None`, `str`, `callable`.
        - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
            given.
        - If `extend_default_variables` was not given as `bool`.
        - If `locked` was not given as `bool`.
        - If `name` was not given as `str`, `iterable` of `str`.
    ValueError
        If a variable name is would be used, what is `module` attribute.
    
    Returns
    -------
    entry_point : `None`, `str`, `callable`
        Plugin specific entry point, to use over the plugin loader's default.
    exit_point : `None`, `str`, `callable`
        Plugin specific exit point, to use over the plugin loader's default.
    extend_default_variables : `bool`
        Whether the plugin should use the loader's default variables or just it's own.
    locked : `bool`
        Whether the given plugin(s) should not be affected by `.{}_all` methods.
    take_snapshot_difference : `bool`
        Whether snapshot feature should be used.
    default_variables : `None`, `HybridValueDictionary` of (`str`, `Any`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If would be empty, is set as `None` instead.
    """
    if not _validate_entry_or_exit(entry_point):
        raise TypeError(
            f'`validate_plugin_parameters` expected `None`, `str` or a `callable` as `entry_point`, got '
            f'{entry_point.__class__.__name__}; {entry_point!r}.'
        )
    
    if not _validate_entry_or_exit(exit_point):
        raise TypeError(
            f'`validate_plugin_parameters` expected `None`, `str` or a `callable` as `exit_point`, got '
            f'{exit_point.__class__.__name__}; {exit_point!r}.'
        )
    
    if variables:
        default_variables = HybridValueDictionary(variables)
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(
                    f'The passed {key!r} is a protected variable name of module type.'
                )
            default_variables[key] = value
    else:
        default_variables = None
    
    extend_default_variables_type = extend_default_variables.__class__
    if extend_default_variables_type is bool:
        pass
    elif issubclass(extend_default_variables_type, int):
        extend_default_variables = bool(extend_default_variables)
    else:
        raise TypeError(
            f'`extend_default_variables` can be `bool`, got {extend_default_variables_type.__name__}; '
            f'{extend_default_variables!r}.'
        )
    
    locked_type = type(locked)
    if locked_type is bool:
        pass
    elif issubclass(locked_type, int):
        locked = bool(locked)
    else:
        raise TypeError(
            f'`locked` can be `bool`, got {locked_type.__name__}; {locked!r}.'
        )
    
    return entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference, default_variables


PROTECTED_NAMES = frozenset((
    '__class__', '__delattr__', '__dict__', '__dir__', '__doc__', '__eq__', '__format__', '__ge__', '__getattribute__',
    '__gt__', '__hash__', '__init__', '__init_subclass__', '__le__', '__lt__', '__module__', '__ne__', '__new__',
    '__reduce__', '__reduce_ex__', '__repr__', '__setattr__', '__sizeof__', '__str__', '__subclasshook__',
    '__weakref__', '_cached', '_set_fileattr', 'cached', 'has_location', 'loader', 'loader_state', 'name', 'origin',
    'parent', 'submodule_search_locations', '__path__', '__spec__'
))



PYTHON_FILE_POSTFIX_NAMES = frozenset(('.py', '.pyd', '.pyc', '.so'))


def _try_get_plugin(plugin_name, plugin_path):
    """
    Tries to get plugin for the given name or path.
    
    Parameters
    ----------
    plugin_name : `None`, `str`
        Plugin's  name.
    plugin_path : `str`
        Path of the plugin file.
    
    Returns
    -------
    plugin : `None`, ``Plugin``
        The found plugin if any.
    """
    try:
        return PLUGIN_LOADER._plugins_by_name[plugin_path]
    except KeyError:
        pass
    
    if plugin_name is None:
        plugin_name = _get_path_plugin_name(plugin_path)
    
    try:
        return PLUGIN_LOADER._plugins_by_name[plugin_name]
    except KeyError:
        pass


def _get_plugin_name_and_path(name, allow_non_existent):
    """
    fetches the name and the path of the first matched plugin. If non is matched raised `ImportError`.
    
    Parameters
    ----------
    name : `str`
        The name to fetch.
    allow_non_existent : `bool`
        Whether non-existent files are allowed.
    
    Raises
    ------
    plugin_name_and_path_pair : `None`, `tuple` (`None`, `str` `str`)
        Plugin's name and path pair.
    
    Raises
    -----
    TypeError
        - If `name` is not `str` nor an `iterable` of `str`.
    """
    if not isinstance(name, str):
        raise TypeError(
            f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
        )
    
    # Return first match
    for plugin_name_and_path_pair in _iter_lookup_plugin_names_and_paths(name, False, True):
        return plugin_name_and_path_pair


def _iter_lookup_plugin_names_and_paths(name, register_directories_as_roots, allow_non_existent):
    """
    Fetches the names and the paths of the given plugin.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    name : `str`
        The name to fetch.
    register_directories_as_roots : `bool`
        Whether directory roots should be registered.
    allow_non_existent : `bool`
        Whether non-existent files are allowed.
    
    Yields
    ------
    plugin_name : `None`, `str`
        Plugin's  name.
    plugin_path : `str`
        Path of the plugin file.
    
    Raises
    ------
    ModuleNotFoundError
        - If `name` could not be detected as an plugin.
    """
    if name.startswith(ABSOLUTE_PATH_PLUGIN_NAME_PREFIX):
        yield None, name
        return
    
    yield from _iter_lookup_path(name, register_directories_as_roots, allow_non_existent)


def _iter_lookup_path(import_name_or_path, register_directories_as_roots, allow_non_existent):
    """
    Detects the root of the given name.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    import_name_or_path : `str`
        An plugin's import name, or it's absolute path.
    register_directories_as_roots : `bool`
        Whether directory roots should be registered.
    allow_non_existent : `bool`
        Whether non-existent files are allowed.
    
    Yields
    ------
    plugin_name : `None`, `str`
        Import name to an plugin file.
    plugin_path : `str`
        Path of the file.
    
    Raise
    -----
    ModuleNotFoundError
        If `import_name_or_path` name could not be detected as a plugin.
    """
    if is_absolute_path(import_name_or_path):
        if exists(import_name_or_path):
            if is_directory(import_name_or_path):
                yield from _iter_lookup_directory(None, import_name_or_path)
                return
            
            if is_file(import_name_or_path):
                yield None, import_name_or_path
                return
        
        if allow_non_existent:
            yield None, import_name_or_path
            return
    
    else:
        path_end = join_paths(*import_name_or_path.split('.'))
        for base_path in sys.path:
            path = join_paths(base_path, path_end)
            if exists(path) and is_directory(path):
                if register_directories_as_roots:
                    register_plugin_root(import_name_or_path)
                yield from _iter_lookup_directory(import_name_or_path, path)
                return
            
            for python_file_postfix_name in PYTHON_FILE_POSTFIX_NAMES:
                file_path = path + python_file_postfix_name
                if exists(file_path) and is_file(file_path):
                    yield import_name_or_path, file_path
                    return
    
    raise ModuleNotFoundError(
        f'The given `import_name_or_path` could not be detected as an plugin nor an absolute path, '
        f'got {import_name_or_path!r}.'
    )


def _iter_lookup_directory(import_name, directory_path):
    """
    Iterates over a directory's import names.
    
    this function is an iterable generator.
    
    Parameters
    ----------
    import_name : `None`, `str`
        The name of the plugin if we would import it.
    directory_path : `str`
        Path to the directory
    
    Yields
    ------
    plugin_name : `None`, `str`
        Detected import names for each applicable file in the directory.
    plugin_path : `str`
        Path of the file.
    """
    for python_file_postfix_name in PYTHON_FILE_POSTFIX_NAMES:
        file_path = join_paths(directory_path, f'__init__{python_file_postfix_name}')
        if exists(file_path) and is_file(file_path):
            yield import_name, file_path
            return
    
    for file_name in list_directory(directory_path):
        if file_name.startswith('.') or (file_name == '__pycache__') or ('tests' in file_name):
            continue
        
        path = join_paths(directory_path, file_name)
        
        if is_file(path):
            for python_file_postfix_name in PYTHON_FILE_POSTFIX_NAMES:
                if file_name.endswith(python_file_postfix_name):
                    if import_name is None:
                        import_name_value = None
                    else:
                        import_name_value = f'{import_name}.{file_name[:-len(python_file_postfix_name)]}'
                    yield import_name_value, path
                    break
            
            continue
        
        if is_directory(path):
            if import_name is None:
                import_name_value = None
            else:
                import_name_value = f'{import_name}.{file_name}'
            yield from _iter_lookup_directory(import_name_value, path)
            continue
        
        # no more cases
        continue


def _get_path_plugin_name(path):
    """
    Creates plugin name from the given path.
    
    Parameter
    ---------
    path : `str`
        Path to a file.
    
    Returns
    -------
    plugin_name : `str`
    """
    file_name = base_name(path)
    if file_name == '__init__.py':
        path = get_parent_directory_path(path)
    else:
        dot_index = file_name.rfind('.')
        if dot_index != -1:
            path = file_name[:dot_index - len(file_name)]
    
    for sys_path in sys.path:
        if path.startswith(sys_path + PATH_SEPARATOR):
            parts = (*path[len(sys_path) + 1:].split(PATH_SEPARATOR),)
            for plugin_root in PLUGIN_ROOTS:
                if is_tuple_starting_with(parts, plugin_root):
                    return '.'.join(parts)
    
    return ABSOLUTE_PATH_PLUGIN_NAME_PREFIX + file_name


def _add_plugin_name_to_plugin_root_names(plugin_root_names, plugin_name):
    """
    Adds the plugin name to the plugin base names.
    
    Parameters
    ----------
    plugin_root_names : `None`, `set` of `str`
        Plugin root names.
    plugin_name : `str`
        The plugin's name.
    
    Returns
    ----------
    plugin_root_names : `set` of `str`
    """
    if plugin_root_names is None:
        plugin_root_names = {plugin_name}
    
    else:
        cut_name = plugin_name
        while True:
            dot_index = cut_name.rfind('.')
            if dot_index == -1:
                plugin_root_names.add(plugin_name)
                break
            
            cut_name = cut_name[:dot_index]
            if cut_name in plugin_root_names:
                break
            
            continue
    
    return plugin_root_names


def _is_plugin_name_in_plugin_root_names(plugin_root_names, plugin_name):
    """
    Returns whether the plugin name is in the given plugin root names.
    
    Parameters
    ----------
    plugin_root_names : `None`, `set` of `str`
        Plugin root names.
    plugin_name : `str`
        The plugin's name.
    
    Returns
    -------
    is_in : `bool`
    """
    if plugin_root_names is None:
        return False
    
    plugin_name_length = len(plugin_name)
    
    for plugin_root_name in plugin_root_names:
        if not plugin_name.startswith(plugin_root_name):
            continue
        
        plugin_root_name_length = len(plugin_root_name)
        if plugin_name_length == plugin_root_name_length:
            return True
        
        if plugin_root_name_length > plugin_name_length and plugin_root_name[plugin_name_length] == '.':
            return True
    
    return False



__all__ = ()

from ..application_command_option_metadata.constants import (
    APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT, APPLICATION_COMMAND_OPTION_MIN_LENGTH_DEFAULT
)

from ..application_command_option_metadata.fields import (
    validate_autocomplete, validate_channel_types, validate_choices, validate_default, validate_max_length,
    validate_max_value, validate_min_length, validate_min_value, validate_options, validate_required
)


FIELDS = {
    'autocomplete': (validate_autocomplete, False),
    'channel_types': (validate_channel_types, None),
    'choices': (validate_choices, None),
    'default': (validate_default, False),
    'max_length': (validate_max_length, APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT),
    'max_value': (validate_max_value, None),
    'min_length': (validate_min_length, APPLICATION_COMMAND_OPTION_MIN_LENGTH_DEFAULT),
    'min_value': (validate_min_value, None),
    'options': (validate_options, None),
    'required': (validate_required, False),
}


def _purge_defaults_and_maybe_raise(keyword_parameters):
    """
    Validates every keyword value. If the output is a default value ignores it.
    
    Raises on extra field and on non-default values.
    
    Parameters
    ----------
    keyword_parameters : `dict` of (`str`, `object)` items
        Keyword parameters passed to an application command option constructor (or copier).
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
        - Extra parameters given.
    ValueError
        - If a parameter's value is incorrect.
        - if a parameter is not applicable for the given option.
    """
    while keyword_parameters:
        key, value = keyword_parameters.popitem()
        
        try:
            validator, default_value = FIELDS[key]
        except KeyError:
            raise TypeError(
                f'Received extra field: field = {key!r}; value = {value!r}.'
            )
        
        validated = validator(value)
        if validated != default_value:
            raise ValueError(
                f'Field not applicable for the option type: field = {key!r}; value = {value!r}.'
            )



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, force_string_parser_factory, nullable_sorted_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, force_string_putter_factory, nullable_string_array_optional_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, force_string_validator_factory, nullable_string_array_validator_factory,
    preinstanced_validator_factory
)

from .constants import TITLE_LENGTH_MAX
from .preinstanced import VerificationScreenStepType

# required

parse_required = bool_parser_factory('required', False)
put_required_into = bool_optional_putter_factory('required', False)
validate_required = bool_validator_factory('required', False)

# title

parse_title = force_string_parser_factory('label')
put_title_into = force_string_putter_factory('label')
validate_title = force_string_validator_factory('title', 0, TITLE_LENGTH_MAX)

# type

parse_type = preinstanced_parser_factory('field_type', VerificationScreenStepType, VerificationScreenStepType.none)
put_type_into = preinstanced_putter_factory('field_type')
validate_type = preinstanced_validator_factory('type', VerificationScreenStepType)

# values

parse_values = nullable_sorted_array_parser_factory('values')
put_values_into = nullable_string_array_optional_putter_factory('values')
validate_values = nullable_string_array_validator_factory('values')



__all__ = ('VerificationScreenStepType',)

from ...bases import Preinstance as P, PreinstancedBase


class VerificationScreenStepType(PreinstancedBase):
    """
    Represents a type of a ``VerificationScreenStep``.

    Attributes
    ----------
    name : `str`
        The verification screen step type's name.
    value : `str`
        The Discord side identifier value of the verification step types.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``VerificationScreenStepType``) items
        Stores the predefined ``VerificationScreenStepType``-s.
    VALUE_TYPE : `type` = `str`
        The verification screen steps' values' type.
    DEFAULT_NAME : `str` = `''`
        The default name of the verification screen step types. New verification screen step types have their name
        generated from their value, so it is not applicable for them.
    
    Every predefined verification screen step type can be accessed as class attribute as well:
    
    +-----------------------+-------------------+-------------------+
    | Class attribute names | Name              | Value             |
    +=======================+===================+===================+
    | none                  | none              | -                 |
    +-----------------------+-------------------+-------------------+
    | rules                 | rules             | TERMS             |
    +-----------------------+-------------------+-------------------+
    | text_input            | text input        | TEXT_INPUT        |
    +-----------------------+-------------------+-------------------+
    | paragraph             | paragraph         | PARAGRAPH         |
    +-----------------------+-------------------+-------------------+
    | multiple_choices      | multiple choices  | MULTIPLE_CHOICES  |
    +-----------------------+-------------------+-------------------+
    | verification          | verification      | VERIFICATION      |
    +-----------------------+-------------------+-------------------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    DEFAULT_NAME = ''
    
    __slots__ = ()
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new verification screen type with the given value.
        
        Parameters
        ----------
        value : `str`
            The verification screen type's identifier value.
        
        Returns
        -------
        self : ``VerificationScreenStepType``
            The verification screen type.
        """
        self = object.__new__(cls)
        self.value = value
        self.name = value.lower().replace('_', ' ')
        self.INSTANCES[value] = self
        return self
    
    
    def __repr__(self):
        """Returns the representation of the verification screen type."""
        return f'{self.__class__.__name__}(value = {self.value!r})'
    
    
    none = P('', 'none')
    rules = P('TERMS', 'rules')
    text_input = P('TEXT_INPUT', 'text input')
    paragraph = P('PARAGRAPH', 'paragraph')
    multiple_choices = P('MULTIPLE_CHOICES', 'multiple_choices')
    verification = P('VERIFICATION', 'verification',)



__all__ = ()

from .fields import (
    put_metadata_values_into, put_platform_name_into, put_platform_user_name_into, validate_metadata_values,
    validate_platform_name, validate_platform_user_name
)


APPLICATION_ROLE_CONNECTION_FIELD_CONVERTERS = {
    'platform_name': (validate_platform_name, put_platform_name_into),
    'platform_user_name': (validate_platform_user_name, put_platform_user_name_into),
    'metadata_values': (validate_metadata_values, put_metadata_values_into),
}




__all__ = ('Embed',)

import warnings

from scarletio import RichAttributeErrorBaseType

from ...utils import sanitize_mentions

from ..embed_author import EmbedAuthor
from ..embed_field import EmbedField
from ..embed_footer import EmbedFooter
from ..embed_image import EmbedImage
from ..embed_thumbnail import EmbedThumbnail
from ..embed_video import EmbedVideo

from .fields import (
    parse_author, parse_color, parse_description, parse_fields, parse_footer, parse_image, parse_provider,
    parse_thumbnail, parse_timestamp, parse_title, parse_type, parse_url, parse_video, put_author_into, put_color_into,
    put_description_into, put_fields_into, put_footer_into, put_image_into, put_provider_into, put_thumbnail_into,
    put_timestamp_into, put_title_into, put_type_into, put_url_into, put_video_into, validate_author, validate_color,
    validate_description, validate_fields, validate_footer, validate_image, validate_provider, validate_thumbnail,
    validate_timestamp, validate_title, validate_type, validate_url, validate_video
)
from .preinstanced import EmbedType


class Embed(RichAttributeErrorBaseType):
    """
    Represents Discord embedded content.
    
    Attributes
    ----------
    author : `None`, ``EmbedAuthor``
        Author information.
    color : `None`, ``Color``
        The color code of the embed. Passing `0` means black, not like at the case of roles.
    description : `None`, `str`
        The main content of the embed.
    fields : `None`, `list` of ``EmbedField``
        Fields' information.
    footer : `None`, ``EmbedFooter``
        Footer information.
    image : `None`, ``EmbedImage``
        Image information.
    provider : `None`, ``EmbedProvider``
        Provider information.
    thumbnail : `None`, ``EmbedThumbnail``
        Thumbnail information.
    timestamp : `None`, `datetime`
        Timestamp of the embed's content. Shows up next to the ``.footer``.
    title : `None`, `str`
        The title of the embed. Shows at the top with intense white characters.
    type : ``EmbedType``
        The type of the embed.Webhook embeds' type must be `EmbedType.rich`.
    url : `None`, `str`
        Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
    video : `None`, ``EmbedVideo``
        Video information.
    
    Examples
    --------
    Example of using local embed file:
    
    ```py
    # Imports
    from hata import Embed, ReuAsyncIO
    
    # Building the embed
    embed = Embed()
    embed.add_image('attachment://image.png')
    
    # Sending the message
    with (await ReuAsyncIO('some_file_path')) as file:
        await client.message_create(channel, embed = embed, file = ('image.png', file))
    ```
    
    Note that you should use async io wrappers, but one which do not closes on `.close` either, but it resets
    itself instead, because if the request fails, the io would be closed and the request could not be done the
    second time.
    """
    __slots__ = (
        'author', 'color', 'description', 'fields', 'footer', 'image', 'provider', 'thumbnail', 'timestamp', 'title',
        'type', 'url', 'video'
    )
    
    def __new__(
        cls,
        title = ...,
        description = ...,
        *extra_positional_parameters, # For deprecations
        type_ = ..., # Deprecated
        author = ...,
        color = ...,
        embed_type = ...,
        fields = ...,
        footer = ...,
        image = ...,
        provider = ...,
        thumbnail = ...,
        timestamp = ...,
        url = ...,
        video = ...,
    ):
        """
        Creates an embed.
        
        Parameters
        ----------
        title : `None`, `str`, Optional
            The title of the embed. Shows at the top with intense white characters.
        description : `None`, `str`, Optional
            The main content of the embed.
        author : `None`, ``EmbedAuthor``, Optional (Keyword only)
            Author information.
        color : `None`, ``Color``, `int`, Optional (Keyword only)
            The color code of the embed. Passing `0` means black, not like at the case of roles.
        embed_type : `EmbedType`, `str`, Optional (Keyword only)
            The type of the embed. Defaults to `EmbedType.rich`.
        fields : `None`, `iterable` of ``EmbedField``, Optional (Keyword only)
            Fields' information.
        footer : `None`, ``EmbedFooter``, Optional (Keyword only)
            Footer information.
        image : `None`, ``EmbedImage``, Optional (Keyword only)
            Image information.
        provider : `None`, ``EmbedProvider``, Optional (Keyword only)
            Provider information.
        thumbnail : `None`, ``EmbedThumbnail``, Optional (Keyword only)
            Thumbnail information.
        timestamp : `None`, `datetime`, Optional (Keyword only)
            Timestamp of the embed's content. Shows up next to the `footer`.
        url : `None`, `str`, Optional (Keyword only)
            Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
        video : `None`, `EmbedVideo`, Optional (Keyword only)
            Video information.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # Handle deprecations, there is a lot!
        if extra_positional_parameters:
            extra_positional_parameter_length = len(extra_positional_parameters)
            if extra_positional_parameter_length >= 4:
                warnings.warn(
                    (
                        f'`embed_type` parameter of `{cls.__name__}.__new__` is moved to be a keyword only '
                        f'parameter and the positional usage is deprecated and will be removed in 2023 August.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
                
                embed_type = extra_positional_parameters[3]
            
            
            if extra_positional_parameter_length >= 3:
                warnings.warn(
                    (
                        f'`timestamp` parameter of `{cls.__name__}.__new__` is moved to be a keyword only '
                        f'parameter and the positional usage is deprecated and will be removed in 2023 August.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
                
                timestamp = extra_positional_parameters[2]
            
            
            if extra_positional_parameter_length >= 2:
                warnings.warn(
                    (
                        f'`url` parameter of `{cls.__name__}.__new__` is moved to be a keyword only '
                        f'parameter and the positional usage is deprecated and will be removed in 2023 August.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
                
                url = extra_positional_parameters[1]
            
            
            if extra_positional_parameter_length >= 1:
                warnings.warn(
                    (
                        f'`color` parameter of `{cls.__name__}.__new__` is moved to be a keyword only '
                        f'parameter and the positional usage is deprecated and will be removed in 2023 August.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
                
                color = extra_positional_parameters[0]
        
        
        if (type_ is not ...):
            warnings.warn(
                (
                    f'`type_` parameter of `{cls.__name__}.__new__` has been renamed to `embed_type`. '
                    f'The option for using `type_` will be removed in 2023 August.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            embed_type = type_
        
        # Validate parameters
        
        # title
        if title is ...:
            title = None
        else:
            title = validate_title(title)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # author
        if author is ...:
            author = None
        else:
            author = validate_author(author)
        
        # color
        if color is ...:
            color = None
        else:
            color = validate_color(color)
        
        # embed_type
        if embed_type is ...:
            embed_type = EmbedType.rich
        else:
            embed_type = validate_type(embed_type)
        
        # fields
        if fields is ...:
            fields = None
        else:
            fields = validate_fields(fields)
        
        # footer
        if footer is ...:
            footer = None
        else:
            footer = validate_footer(footer)
        
        # image
        if image is ...:
            image = None
        else:
            image = validate_image(image)
        
        # provider
        if provider is ...:
            provider = None
        else:
            provider = validate_provider(provider)
        
        # thumbnail
        if thumbnail is ...:
            thumbnail = None
        else:
            thumbnail = validate_thumbnail(thumbnail)
        
        # timestamp
        if timestamp is ...:
            timestamp = None
        else:
            timestamp = validate_timestamp(timestamp)
        
        # url
        if url is ...:
            url = None
        else:
            url = validate_url(url)
        # video
        if video is ...:
            video = None
        else:
            video = validate_video(video)
        
        self = object.__new__(cls)
        self.author = author
        self.color = color
        self.description = description
        self.fields = fields
        self.footer = footer
        self.image = image
        self.provider = provider
        self.thumbnail = thumbnail
        self.timestamp = timestamp
        self.title = title
        self.type = embed_type
        self.url = url
        self.video = video
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an embed object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embed data received from Discord.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.author = parse_author(data)
        self.color = parse_color(data)
        self.description = parse_description(data)
        self.fields = parse_fields(data)
        self.footer = parse_footer(data)
        self.image = parse_image(data)
        self.provider = parse_provider(data)
        self.thumbnail = parse_thumbnail(data)
        self.timestamp = parse_timestamp(data)
        self.title = parse_title(data)
        self.type = parse_type(data)
        self.url = parse_url(data)
        self.video = parse_video(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the embed core to json serializable `dict` representing it.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_author_into(self.author, data, defaults, include_internals = include_internals)
        put_color_into(self.color, data, defaults)
        put_description_into(self.description, data, defaults)
        put_fields_into(self.fields, data, defaults)
        put_footer_into(self.footer, data, defaults, include_internals = include_internals)
        put_image_into(self.image, data, defaults, include_internals = include_internals)
        put_thumbnail_into(self.thumbnail, data, defaults, include_internals = include_internals)
        put_timestamp_into(self.timestamp, data, defaults)
        put_title_into(self.title, data, defaults)
        put_type_into(self.type, data, defaults)
        put_url_into(self.url, data, defaults)
        
        if include_internals:
            put_provider_into(self.provider, data, defaults, include_internals = include_internals)
            put_video_into(self.video, data, defaults, include_internals = include_internals)
        
        return data
    
    
    def clear(self):
        """
        Clears the embed.
        """
        self.author = None
        self.color = None
        self.description = None
        self.fields = None
        self.footer = None
        self.image = None
        self.provider = None
        self.thumbnail = None
        self.timestamp = None
        self.title = None
        self.type = None
        self.url = None
        self.video = None
    
    
    def _update_sizes(self, data):
        """
        Updates the size information of the embed.
        
        Called when a ``Message`` is edited, but no `edited` timestamp is included with the data. Returns `0` if
        received data does not contain images, if does, then `1`.
        
        This method tries to update the embed's `image`, `thumbnail` amd `video` with their sizes. If any of those is
        not set already (for any reason), then it also creates them.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embed data received from Discord.
        
        Returns
        -------
        changed : `int`
        """
        changed = 0
        try:
            image_data = data['image']
        except KeyError:
            pass
        else:
            image = self.image
            if image is None:
                self.image = EmbedImage.from_data(image_data)
            else:
                image.height = image_data.get('height', 0)
                image.width = image_data.get('width', 0)
            changed = 1
        
        try:
            thumbnail_data = data['thumbnail']
        except KeyError:
            pass
        else:
            thumbnail = self.thumbnail
            if thumbnail is None:
                self.thumbnail = EmbedThumbnail.from_data(thumbnail_data)
            else:
                thumbnail.height = thumbnail_data.get('height', 0)
                thumbnail.width = thumbnail_data.get('width', 0)
            changed = 1
        
        try:
            video_data = data['video']
        except KeyError:
            pass
        else:
            video = self.video
            if video is None:
                self.video = EmbedVideo.from_data(video_data)
            else:
                video.height = video_data.get('height', 0)
                video.width = video_data.get('width', 0)
            changed = 1
        
        return changed
    
    
    def _set_sizes(self, data):
        """
        Updates the size information of the embed.
        
        Familiar to ``._update_sizes`` but it not checks whether the embed's images change, just updates them.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embed data received from Discord.
        """
        try:
            image_data = data['image']
        except KeyError:
            pass
        else:
            image = self.image
            if image is None:
                self.image = EmbedImage.from_data(image_data)
            else:
                image.height = image_data.get('height', 0)
                image.width = image_data.get('width', 0)
        
        try:
            thumbnail_data = data['thumbnail']
        except KeyError:
            pass
        else:
            thumbnail = self.thumbnail
            if thumbnail is None:
                self.thumbnail = EmbedThumbnail.from_data(thumbnail_data)
            else:
                thumbnail.height = thumbnail_data.get('height', 0)
                thumbnail.width = thumbnail_data.get('width', 0)
        
        try:
            video_data = data['video']
        except KeyError:
            pass
        else:
            video = self.video
            if video is None:
                self.video = EmbedVideo.from_data(video_data)
            else:
                video.height = video_data.get('height', 0)
                video.width = video_data.get('width', 0)
    
    
    def __len__(self):
        """Returns the embed's contents' length."""
        length = 0
        
        for content in self.iter_contents():
            length += len(content)
        
        return length
    
    
    def __bool__(self):
        """Returns whether the embed is not empty."""
        # author
        author = self.author
        if (author is not None) and author:
            return True
        
        # color
        if (self.color is not None):
            return True
        
        # description
        description = self.description
        if (description is not None) and description:
            return True
        
        # fields
        for field in self.iter_fields():
            if field:
                return True
        
        # footer
        footer = self.footer
        if (footer is not None) and footer:
            return True
        
        # image
        image = self.image
        if (image is not None) and image:
            return True
        
        # provider
        provider = self.provider
        if (provider is not None) and provider:
            return True
        
        # thumbnail
        thumbnail = self.thumbnail
        if (thumbnail is not None) and thumbnail:
            return True
        
        # timestamp
        if (self.timestamp is not None):
            return True
        
        # title
        title = self.title
        if (title is not None) and title:
            return True
        
        # type
        # Not applicable
        
        # url
        if (self.url is not None):
            return True
        
        # video
        video = self.video
        if (video is not None) and video:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the representation of the embed."""
        return f'<{self.__class__.__name__} length = {len(self)}>'
    
    
    def __eq__(self, other):
        """Returns whether the two embeds are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # author
        if self.author != other.author:
            return False
        
        # color
        if self.color != other.color:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # fields
        if self.fields != other.fields:
            return False
        
        # footer
        if self.footer != other.footer:
            return False
        
        # image
        if self.image != other.image:
            return False
        
        # provider
        if self.provider != other.provider:
            return False
        
        # thumbnail
        if self.thumbnail != other.thumbnail:
            return False
        
        # timestamp
        if self.timestamp != other.timestamp:
            return False
        
        # title
        if self.title != other.title:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        # video
        if self.video != other.video:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the embed's hash value."""
        hash_value = 0
        
        # author
        author = self.author
        if (author is not None):
            hash_value ^= 1 << 0
            hash_value ^= hash(author)
        
        # color
        color = self.color
        if (color is not None):
            hash_value ^= 1 << 1
            hash_value ^= color
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= 1 << 2
            hash_value ^= hash(description)
        
        # fields
        fields = self.fields
        if (fields is not None):
            hash_value ^= len(fields) << 3
            
            for field in fields:
                hash_value ^= hash(field)
        
        # footer
        footer = self.footer
        if (footer is not None):
            hash_value ^= 1 << 8
            hash_value ^= hash(footer)
        
        # image
        image = self.image
        if (image is not None):
            hash_value ^= 1 << 9
            hash_value ^= hash(image)
        
        # provider
        provider = self.provider
        if (provider is not None):
            hash_value ^= 1 << 10
            hash_value ^= hash(provider)
        
        # thumbnail
        thumbnail = self.thumbnail
        if (thumbnail is not None):
            hash_value ^= 1 << 11
            hash_value ^= hash(thumbnail)
        
        # timestamp
        timestamp = self.timestamp
        if (timestamp is not None):
            hash_value ^= 1 << 12
            hash_value ^= hash(timestamp)
        
        # title
        title = self.title
        if (title is not None):
            hash_value ^= 1 << 13
            hash_value ^= hash(title)
        
        # type
        hash_value ^= hash(self.type)
        
        # url
        url = self.url
        if (url is not None):
            hash_value ^= 1 << 14
            hash_value ^= hash(url)
        
        # video
        video = self.video
        if (video is not None):
            hash_value ^= 1 << 15
            hash_value ^= hash(video)
        
        return hash_value
    
    
    def clean_copy(self, guild = None):
        """
        Creates a clean copy of the embed by removing the mentions in it's contents.
        
        Parameters
        ----------
        guild : `None`, ``Guild`` = `None`, Optional
            The respective guild as a context to look up guild specific names of entities.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        author = self.author
        if (author is not None):
            author = author.clean_copy(guild)
        new.author = author
        
        new.color = self.color
        new.description = sanitize_mentions(self.description, guild)
        
        fields = self.fields
        if (fields is not None):
            fields = [field.clean_copy(guild) for field in fields]
        new.fields = fields
        
        footer = self.footer
        if (footer is not None):
            footer = footer.clean_copy(guild)
        new.footer = footer
        
        image = self.image
        if (image is not None):
            image = image.clean_copy(guild)
        new.image = image
        
        provider = self.provider
        if (provider is not None):
            provider = provider.clean_copy(guild)
        new.provider = provider
        
        thumbnail = self.thumbnail
        if (thumbnail is not None):
            thumbnail = thumbnail.clean_copy(guild)
        new.thumbnail = thumbnail
        
        new.timestamp = self.timestamp
        new.title = sanitize_mentions(self.title, guild)
        new.type = self.type
        new.url = self.url
        
        video = self.video
        if (video is not None):
            video = video.clean_copy(guild)
        new.video = video
        
        return new
    
    
    def copy(self):
        """
        Copies the embed.
        
        Returns
        -------
        new : `instance<type<self>`
        """
        new = object.__new__(type(self))
        
        author = self.author
        if (author is not None):
            author = author.copy()
        new.author = author
        
        new.color = self.color
        new.description = self.description
        
        fields = self.fields
        if (fields is not None):
            fields = [field.copy() for field in fields]
        new.fields = fields
        
        footer = self.footer
        if (footer is not None):
            footer = footer.copy()
        new.footer = footer
        
        image = self.image
        if (image is not None):
            image = image.copy()
        new.image = image
        
        provider = self.provider
        if (provider is not None):
            provider = provider.copy()
        new.provider = provider
        
        thumbnail = self.thumbnail
        if (thumbnail is not None):
            thumbnail = thumbnail.copy()
        new.thumbnail = thumbnail
        
        new.timestamp = self.timestamp
        new.title = self.title
        new.type = self.type
        new.url = self.url
        
        video = self.video
        if (video is not None):
            video = video.copy()
        new.video = video
        
        return new
    
    
    def copy_with(
        self,
        *,
        author = ...,
        color = ...,
        description = ...,
        embed_type = ...,
        fields = ...,
        footer = ...,
        image = ...,
        provider = ...,
        thumbnail = ...,
        timestamp = ...,
        title = ...,
        url = ...,
        video = ...
    ):
        """
        Copies the embed with the given parameters.
        
        Parameters
        ----------
        author : `None`, ``EmbedAuthor``, Optional (Keyword only)
            Author information.
        color : `None`, ``Color``, `int`, Optional (Keyword only)
            The color code of the embed. Passing `0` means black, not like at the case of roles.
        description : `None`, `str`, Optional (Keyword only)
            The main content of the embed.
        embed_type : ``EmbedType``, `str`, Optional (Keyword only)
            The type of the embed. Webhook embed's type must be `EmbedType.rich`.
        fields : `None`, `iterable` of ``EmbedField``, Optional (Keyword only)
            Fields' information.
        footer : `None`, ``EmbedFooter``, Optional (Keyword only)
            Footer information.
        image : `None`, ``EmbedImage``, Optional (Keyword only)
            Image information.
        provider : `None`, ``EmbedProvider``, Optional (Keyword only)
            Provider information.
        thumbnail : `None`, ``EmbedThumbnail``, Optional (Keyword only)
            Thumbnail information.
        timestamp : `None`, `datetime`, Optional (Keyword only)
            Timestamp of the embed's content. Shows up next to the ``.footer``.
        title : `None`, `str`, Optional (Keyword only)
            The title of the embed. Shows at the top with intense white characters.
        url : `None`, `str`, Optional (Keyword only)
            Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
        video : `None`, `EmbedVideo`, Optional (Keyword only)
            Video information.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # author
        if author is ...:
            author = self.author
            if (author is not None):
                author = author.copy()
        else:
            author = validate_author(author)
        
        # color
        if color is ...:
            color = self.color
        else:
            color = validate_color(color)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # fields
        if fields is ...:
            fields = self.fields
            if (fields is not None):
                fields = [field.copy() for field in fields]
        else:
            fields = validate_fields(fields)
        
        # footer
        if footer is ...:
            footer = self.footer
            if (footer is not None):
                footer = footer.copy()
        else:
            footer = validate_footer(footer)
        
        # image
        if image is ...:
            image = self.image
            if (image is not None):
                image = image.copy()
        else:
            image = validate_image(image)
        
        # provider
        if provider is ...:
            provider = self.provider
            if (provider is not None):
                provider = provider.copy()
        else:
            provider = validate_provider(provider)
        
        # thumbnail
        if thumbnail is ...:
            thumbnail = self.thumbnail
            if (thumbnail is not None):
                thumbnail = thumbnail.copy()
        else:
            thumbnail = validate_thumbnail(thumbnail)
        
        # timestamp
        if timestamp is ...:
            timestamp = self.timestamp
        else:
            timestamp = validate_timestamp(timestamp)
        
        # title
        if title is ...:
            title = self.title
        else:
            title = validate_title(title)
        
        # type
        if embed_type is ...:
            embed_type = self.type
        else:
            embed_type = validate_type(embed_type)
        
        # url
        if url is ...:
            url = self.url
        else:
            url = validate_url(url)
        
        # video
        if video is ...:
            video = self.video
            if (video is not None):
                video = video.copy()
        else:
            video = validate_video(video)
        
        new = object.__new__(type(self))
        new.author = author
        new.color = color
        new.description = description
        new.fields = fields
        new.footer = footer
        new.image = image
        new.provider = provider
        new.thumbnail = thumbnail
        new.timestamp = timestamp
        new.title = title
        new.type = embed_type
        new.url = url
        new.video = video
        return new
    
    # Extra Utility
    
    def iter_fields(self):
        """
        Iterates over the fields of the embed.
        
        This method is an iterable generator.
        
        Yields
        ------
        field : ``EmbedField``
        """
        fields = self.fields
        if (fields is not None):
            yield from fields
    
    
    def iter_contents(self):
        """
        Iterates over the embed's contents.
        
        This method is an iterable generator.
        
        Yields
        -------
        contents : `str`
        """
        provider = self.provider
        if (provider is not None):
            yield from provider.iter_contents()
        
        author = self.author
        if (author is not None):
            yield from author.iter_contents()
        
        title = self.title
        if (title is not None):
            yield title
        
        description = self.description
        if (description is not None):
            yield description
        
        for field in self.iter_fields():
            yield from field.iter_contents()
        
        footer = self.footer
        if (footer is not None):
            yield from footer.iter_contents()
    
    
    @property
    def contents(self):
        """
        Returns the embed's contents.
        
        The embed's contents are the following:
        - `.author.name`
        - `.description`
        - `.fields[n].name`
        - `.fields[n].value`
        - `.title`
        - `.footer.text`
        - `.provider.name`
        
        Returns
        -------
        contents : `list` of `str`
        """
        return [*self.iter_contents()]
    
    # Field adder methods

    def add_author(self, name = None, icon_url = None, url = None):
        """
        Adds an ``EmbedAuthor`` to the embed with the given parameters.
        
        Parameters
        ----------
        name : `None`, `str` = `None`, Optional
            The name of the author.
        icon_url : `None`, `str` = `None`, Optional
            An url of the author's icon. Can be http(s) or attachment.
        url : `None`, `str` = `None`, Optional
            The url of the author.
        
        Returns
        -------
        self : `self`
        """
        self.author = EmbedAuthor(name, icon_url, url)
        return self
    
    # fields

    def add_field(self, name, value, inline = False):
        """
        Adds an ``EmbedField`` to the end of the embed's fields.
        
        Parameters
        ----------
        name : `str`
            The name of the field.
        value : `str`
            The value of the field.
        inline : `bool` = `False`, Optional
            Whether this field should display inline.
        
        Returns
        -------
        self : `self`
        """
        field = EmbedField(name, value, inline)
        
        fields = self.fields
        if fields is None:
            fields = []
            self.fields = fields
        
        fields.append(field)
        return self
    
    
    def insert_field(self, index, name, value, inline = False):
        """
        Inserts an ``EmbedField`` to the embed's fields at the specified `index`.
        
        Parameters
        ----------
        index : `int`
            The index before the field should be inserted.
        name : `str`
            The name of the field.
        value : `str`
            The value of the field.
        inline : `bool` = `False`, Optional
            Whether this field should display inline.
        
        Returns
        -------
        self : `self`
        """
        field = EmbedField(name, value, inline)
        
        fields = self.fields
        if fields is None:
            fields = []
            self.fields = fields
        
        fields.insert(index, field)
        return self
    
    
    def get_field(self, index):
        """
        Returns the embed's field at the given `index`.
        
        Parameters
        ----------
        index : `int`
            The index of the field.
        
        Returns
        -------
        embed_field : ``EmbedField``
        
        Raises
        ------
        IndexError
            Index out of the fields' range.
        """
        fields = self.fields
        if fields is None:
            raise IndexError(
                f'Index out of the fields\' range, got index = {index!r}; length = 0.'
            ) from None
        
        try:
            field = fields[index]
        except IndexError:
            raise IndexError(
                f'Index out of the fields\' range, got index = {index!r}; length = {len(fields)!r}.'
            ) from None
        
        return field
    
    
    def append_field(self, field):
        """
        Appends the embed's fields with the given field.
        
        Parameters
        ----------
        field : ``EmbedField``
            The field to append the embed's field with.
        """
        if not isinstance(field, EmbedField):
            raise TypeError(
                f'`field` can be `{EmbedField.__name__}`, got {field.__class__.__name__}; {field!r}.'
            )
        
        fields = self.fields
        if fields is None:
            fields = []
            self.fields = fields
        
        fields.append(field)
    
    
    def set_field(self, index, field):
        """
        Sets the given ``EmbedField`` at the given `index` to the fields of the embed.
        
        Parameters
        ----------
        index : `int`
            The index of the field to set.
        field : ``EmbedField``
            The embed field to set at the given `index`.
        
        Raises
        ------
        IndexError
            Index out of the fields' range.
        """
        if not isinstance(field, EmbedField):
            raise TypeError(
                f'`field` can be `{EmbedField.__name__}`, got {field.__class__.__name__}; {field!r}.'
            )
        
        fields = self.fields
        if fields is None:
            raise IndexError(
                f'Index out of the fields\' range, got index={index!r}; length = 0.'
            ) from None
        
        
        try:
            fields[index] = field
        except IndexError:
            raise IndexError(
                f'Index out of the fields\' range, got index={index!r}; length = {len(fields)!r}.'
            ) from None
    
    
    def del_field(self, index):
        """
        Removes the field of the embed on the given index.
        
        Parameters
        ----------
        index : `int`
            The index of the field to remove.
        
        Raises
        ------
        IndexError
            Index out of the fields' range.
        """
        fields = self.fields
        if fields is None:
            raise IndexError(
                f'Index out of the fields\' range, got index = {index!r}; length = 0.'
            )
        
        try:
            del fields[index]
        except IndexError:
            raise IndexError(
                f'Index out of the fields\' range, got index = {index!r}; length = {len(fields)!r}.'
            ) from None
        
        if not fields:
            self.fields = None
    
    remove_field = del_field
    
    # footer
    
    def add_footer(self, text, icon_url = None):
        """
        Adds an ``EmbedFooter`` to the embed with the given parameters.
        
        Parameters
        ----------
        text : `str`
            The footer's text.
        icon_url : `None`, `str` = `None`, Optional
            An url of the footer's icon. Can be http(s) or attachment.
        
        Returns
        -------
        self : `self`
        """
        self.footer = EmbedFooter(text, icon_url)
        return self
    
    # image
    
    def add_image(self, url):
        """
        Adds an ``EmbedImage`` to the embed with the given `url`.
        
        Parameters
        ----------
        url : `str`
            The url of the image. Can be http(s) or attachment.
        
        Returns
        -------
        self : `self`
        """
        self.image = EmbedImage(url)
        return self
    
    # thumbnail
    
    def add_thumbnail(self, url):
        """
        Adds an ``EmbedThumbnail`` to the embed with the given `url`.
        
        Parameters
        ----------
        url : `str`
            The url of the thumbnail. Can be http(s) or attachment.
        
        Returns
        -------
        self : `self`
        """
        self.thumbnail = EmbedThumbnail(url)
        return self
    
    
    @property
    def _data(self):
        """
        Deprecated and will be removed in 2023 August.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}._data` is deprecated, and will be removed in 2023 August. '
                f'Please use `.to_data()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.to_data(include_internals = True)
    
    
    @_data.setter
    def _data(self, data):
        warnings.warn(
            (
                f'`{self.__class__.__name__}._data` is deprecated, and will be removed in 2023 August. '
                f'Please use `.from_data(data)` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        self.author = parse_author(data)
        self.color = parse_color(data)
        self.description = parse_description(data)
        self.fields = parse_fields(data)
        self.footer = parse_footer(data)
        self.image = parse_image(data)
        self.provider = parse_provider(data)
        self.thumbnail = parse_thumbnail(data)
        self.timestamp = parse_timestamp(data)
        self.title = parse_title(data)
        self.type = parse_type(data)
        self.url = parse_url(data)
        self.video = parse_video(data)
